import logging

import torch
from torch import nn

from ..nn import LukasiewiczChannelOrBlock, LukasiewiczChannelAndBlock, Predicates


class BanditNRNModule(nn.Module):

    def __init__(
            self,
            input_size: int,
            output_size: int,
            layer_sizes: list,
            feature_names: list,
            n_selected_features_input: int,
            n_selected_features_internal: int,
            n_selected_features_output: int,
            perform_prune_quantile: float,
            ucb_scale: float,
            normal_form: str = 'dnf',
            add_negations: bool = False,
            weight_init: float = 0.2,
            logits: bool = False
    ):
        """
        Initialize a Bandit Reinforced Neural Reasoning Network module.

        Args:
            input_size (int): number of features from input.
            output_size (int): number of outputs.
            layer_sizes (list): A list containing the number of output logics for each layer.
            feature_names (list): A list of feature names.
            n_selected_features_input (int): The number of features to include in each logic in the input layer.
            n_selected_features_internal (int): The number of logics to include in each logic in the internal layers.
            n_selected_features_output (int): The number of logics to include in each logic in the output layer.
            perform_prune_quantile (float): The quantile to use for pruning randomized RN.
            ucb_scale (float): The scale of the confidence interval in the multi-armed bandit policy.
                               c = 1.96 is a 95% confidence interval.
            normal_form (str): 'dnf' for disjunctive normal form network; 'cnf' for conjunctive normal form network.
            add_negations (bool): add negations of logic.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
        """
        super(BanditNRNModule, self).__init__()
        self.ucb_scale = ucb_scale
        self.input_size = input_size
        self.output_size = output_size
        self.layer_sizes = layer_sizes
        self.feature_names = feature_names
        self.n_selected_features_input = n_selected_features_input
        self.n_selected_features_internal = n_selected_features_internal
        self.n_selected_features_output = n_selected_features_output
        self.perform_prune_quantile = perform_prune_quantile
        self.normal_form = normal_form
        self.add_negations = add_negations
        self.weight_init = weight_init
        self.logits = logits
        self.logger = logging.getLogger(self.__class__.__name__)

        assert self.normal_form in ['cnf', 'dnf'], "'normal_form' must be one of 'dnf', 'cnf'."

        # input layer
        if self.normal_form == 'dnf':
            input_layer = LukasiewiczChannelAndBlock(
                channels=output_size,
                in_features=input_size,
                out_features=layer_sizes[0],
                n_selected_features=n_selected_features_input,
                parent_weights_dimension='out_features',
                operands=Predicates(feature_names=feature_names),
                outputs_key='0',
                add_negations=add_negations,
                weight_init=weight_init
            )
        else:
            input_layer = LukasiewiczChannelOrBlock(
                channels=output_size,
                in_features=input_size,
                out_features=layer_sizes[0],
                n_selected_features=n_selected_features_input,
                parent_weights_dimension='out_features',
                operands=Predicates(feature_names=feature_names),
                outputs_key='0',
                add_negations=add_negations,
                weight_init=weight_init
            )

        model_layers = [input_layer]

        # add alternating internal layers
        for i in range(1, len(layer_sizes)):

            operands = model_layers[-1]

            if i % 2 == 0:
                if self.normal_form == 'dnf':
                    internal_layer = LukasiewiczChannelAndBlock(
                        channels=output_size,
                        in_features=layer_sizes[i - 1],
                        out_features=layer_sizes[i],
                        n_selected_features=n_selected_features_internal,
                        parent_weights_dimension='out_features',
                        operands=operands,
                        outputs_key=str(i),
                        weight_init=weight_init
                    )
                else:
                    internal_layer = LukasiewiczChannelOrBlock(
                        channels=output_size,
                        in_features=layer_sizes[i],
                        out_features=layer_sizes[i],
                        n_selected_features=n_selected_features_internal,
                        parent_weights_dimension='out_features',
                        operands=operands,
                        outputs_key=str(i),
                        weight_init=weight_init
                    )

            else:
                if self.normal_form == 'dnf':
                    internal_layer = LukasiewiczChannelOrBlock(
                        channels=output_size,
                        in_features=layer_sizes[i],
                        out_features=layer_sizes[i],
                        n_selected_features=n_selected_features_internal,
                        parent_weights_dimension='out_features',
                        operands=operands,
                        outputs_key=str(i),
                        weight_init=weight_init
                    )
                else:
                    internal_layer = LukasiewiczChannelAndBlock(
                        channels=output_size,
                        in_features=layer_sizes[i - 1],
                        out_features=layer_sizes[i],
                        n_selected_features=n_selected_features_internal,
                        parent_weights_dimension='out_features',
                        operands=operands,
                        outputs_key=str(i),
                        weight_init=weight_init
                    )

            model_layers += [internal_layer]

        # OR LAYER
        if len(layer_sizes) % 2 == 0:
            if self.normal_form == 'dnf':
                self.output_layer = LukasiewiczChannelAndBlock(
                    channels=output_size,
                    in_features=layer_sizes[-1],
                    out_features=1,
                    n_selected_features=n_selected_features_output,
                    parent_weights_dimension='out_features',
                    operands=model_layers[-1],
                    outputs_key='output_layer',
                    weight_init=weight_init
                )
            else:
                self.output_layer = LukasiewiczChannelOrBlock(
                    channels=output_size,
                    in_features=layer_sizes[-1],
                    out_features=1,
                    n_selected_features=n_selected_features_output,
                    parent_weights_dimension='out_features',
                    operands=model_layers[-1],
                    outputs_key='output_layer',
                    weight_init=weight_init
                )
        else:
            if self.normal_form == 'dnf':
                self.output_layer = LukasiewiczChannelOrBlock(
                    channels=output_size,
                    in_features=layer_sizes[-1],
                    out_features=1,
                    n_selected_features=n_selected_features_output,
                    parent_weights_dimension='out_features',
                    operands=model_layers[-1],
                    outputs_key='output_layer',
                    weight_init=weight_init
                )
            else:
                self.output_layer = LukasiewiczChannelAndBlock(
                    channels=output_size,
                    in_features=layer_sizes[-1],
                    out_features=1,
                    n_selected_features=n_selected_features_output,
                    parent_weights_dimension='out_features',
                    operands=model_layers[-1],
                    outputs_key='output_layer',
                    weight_init=weight_init
                )

        self.model = torch.nn.Sequential(*model_layers)

        self.logger = logging.getLogger(self.__class__.__name__)

    def forward(self, x: torch.Tensor):
        x = self.model(x)
        if self.logits:
            return torch.special.logit(self.output_layer(x).squeeze(-1).squeeze(-1), eps=1e-6)
        else:
            return self.output_layer(x).squeeze(-1).squeeze(-1)



import numpy as np

# constants
GREATER_SIGNS = [">", ">=", "greater", "greater than", "greater than or equal", "greater than or equal to"]
GREATER_OR_EQUAL_SIGNS = [">=", "greater than or equal", "greater than or equal to"]
LESS_SIGNS = ["<", "<=", "less", "less than", "less than or equal", "less than or equal to"]
LESS_OR_EQUAL_SIGNS = ["<=", "less than or equal", "less than or equal to"]


class Node:
    def __init__(self, explanation, level, logic_type, print_type: str, importance_weight=np.nan, feature_importance=False):
        self.explanation = explanation.replace('"', "")
        self.level = level
        self.type = logic_type  # "AND", "OR", "NOT" or "LEAF"
        self.operands = []
        self.print_type = print_type # "logical", "logical-natural" or "natural" 
        assert logic_type in ["AND", "OR", "NOT", "LEAF"], "`logic_type` must be one of 'AND', 'OR', 'NOT', 'LEAF'."

        if self.type == "AND":
            self.natural_type = "all the following are true:"
        elif self.type == "OR":
            self.natural_type = "any of the following are true:"
        elif self.type == "NOT":
            self.natural_type = "it was NOT true that"
        elif self.type == "LEAF":
            self.natural_type = None

        self.feature_importance = feature_importance
        self.importance_weight = 1.
        if self.feature_importance:
            self.importance_weight = importance_weight

    def parse_explanation(self):
        self.feature, self.sign, self.value = Explanation.parse_explanation(self.explanation)
        if len(self.value) > 0:
            self.explanation = self.feature + " was " + self.sign + " " + str(self.value)
            self.explanation = self.explanation.replace("  ", " ")
        
    def add_operands(self, nodes):
        self._add_operand(nodes, 0)

    def _add_operand(self, nodes, i):
        for x in nodes[i:]:
            if x.level <= self.level:
                break

            if x.level == self.level + 1:  # is the node a operand
                self.operands.append(x)
                i = x._add_operand(nodes, i + 1)

        return i

    def print_tree(self):
        indentation = "\t" * self.level
        print(f"{indentation}{str(self)}")
        if self.operands:
            for operand in self.operands:
                operand.print_tree()
                
    def tree_to_string(self, input_string=''):
        indentation = "\t" * self.level
        if self.type == 'LEAF':
            self.parse_explanation()
        output_string = input_string + (f"\n{indentation}{str(self)}")
        if self.operands:
            for operand in self.operands:
                output_string = operand.tree_to_string(output_string)
        return output_string

    def decrease_level(self):
        self.level -= 1
        if self.operands:
            for operand in self.operands:
                operand.decrease_level()
        
    def simplify_and_and(self):
        total_operands = list()
        if not (self.type == "AND"):
            return 
        for child in self.operands:
            if not child.type == "AND":
                total_operands.append(child)
            else:
                for grandchild in child.operands:
                    grandchild.decrease_level()
                    total_operands.append(grandchild)
        self.operands = total_operands
          
    def simplify_or_or(self):
        total_operands = list() 
        if not (self.type == "OR"):
            return 
        for child in self.operands:
            if not child.type == "OR":
                total_operands.append(child)
            else:
                for grandchild in child.operands:
                    grandchild.decrease_level()
                    total_operands.append(grandchild)
        self.operands = total_operands
        
    def simplify_not_and(self):
        if not ((self.type == "NOT") and (len(self.operands)==1) and (self.operands[0].type == "AND")):
            return
        self.type = "OR"
        and_node = self.operands[0]
        self.explanation = and_node.explanation
        self.operands = [Explanation._negate_and_decrease_level(operand) for operand in and_node.operands]
        return

    def simplify_not_or(self):
        if not ((self.type == "NOT") and (len(self.operands)==1) and (self.operands[0].type == "OR")):
            return
        self.type = "AND"
        or_node = self.operands[0]
        self.explanation = or_node.explanation
        self.operands = [Explanation._negate_and_decrease_level(operand) for operand in or_node.operands]
        return
    
    def collapse_not(self):
        if not ((self.type == "NOT") and (len(self.operands)==1) and (self.operands[0].type == "LEAF")):
            return
        self.explanation = self.operands[0].explanation
        self.operands = []
        return
    
    def negate_explanation(self, substr1, substr2):
        if (len(self.operands) > 0) or (self.type != "NOT"):
            return
        if not substr1 in self.explanation:
            return
        self.explanation = self.explanation.replace(substr1, substr2)
        self.parse_explanation()    
        self.type = "LEAF"
        return
    
    def convert_not_to_leaf(self):
        self.collapse_not()
        self.negate_explanation("less than or equal to", "greater than")
        self.negate_explanation("less than", "greater than or equal to")
        self.negate_explanation("greater than or equal to", "less than")
        self.negate_explanation("greater than", "less than or equal to")
        self.negate_explanation(" <= ", " > ")
        self.negate_explanation(" >= ", " < ") 
        self.negate_explanation(" < ", " >= ")
        self.negate_explanation(" > ", " <= ")
        return
     
    def delete_duplicated_nodes(self):
        self.operands = list(set(self.operands))
        return
    
    def sort_operands(self):
        for child in self.operands:
            child.sort_operands()
        self.operands = Explanation.sort_nodes(self.operands)
        return 

    def get_feature_importances(self, parents_importance=1., importances_dict=None):
        if importances_dict is None:
            importances_dict = dict() 
        importance_weight = parents_importance * self.importance_weight
        if self.type == "LEAF":
            if self.explanation in importances_dict:
                importances_dict[self.explanation].append(importance_weight)
            else:
                importances_dict[self.explanation] = [importance_weight]
        else:
            for child in self.operands:
                child.get_feature_importances(importance_weight, importances_dict)
        return importances_dict
    
    def __eq__(self, other):
        if not isinstance(other, Node):
            return False
        if (len(self.operands) < 1) and (len(other.operands) < 1) and \
            (self.level == other.level) and (self.explanation == other.explanation) and \
            (self.type == other.type):
            return True
        elif (len(self.operands) < 1) and (len(other.operands) < 1):
            return False
        return set(self.operands) == set(other.operands)

    def __str__(self) -> str:
        if self.feature_importance:
            if self.type == "LEAF":
                return f"{self.explanation[0].capitalize() + self.explanation[1:] + ' [' + str(self.importance_weight) + ']'}"
            if self.print_type == 'logical-natural':
                return f"{self.natural_type.capitalize()} {self.explanation}"
            return f"{self.type} {self.explanation} [{str(self.importance_weight)}]"
        else:
            if self.type == "LEAF":
                return f"{self.explanation[0].capitalize() + self.explanation[1:]}"
            if self.print_type == 'logical-natural':
                return f"{self.natural_type.capitalize()} {self.explanation}"
            return f"{self.type} {self.explanation}"

    def __repr__(self):
        return "Node(%s, %d, %s, %s)" % (self.explanation, self.level, self.type, self.print_type)
    
    def __hash__(self):
        return hash(self.__repr__())


class Explanation:
    def __init__(self, input_string, print_type):
        self.root = Explanation._text_to_tree(input_string, print_type)

    @staticmethod
    def _text_to_phrases(text):
        phrases = []
        phrase = ''
        for c in text:
            if not c in ['(', ')', ',']:
                phrase += c
            else:
                phrases.append(phrase)
                phrases.append(c)
                phrase = ''
        if not c in ['(', ')', ',']: # in case if there are no parentheses in the text
            phrases.append(phrase)
        phrases = [phrase.strip() for phrase in phrases]
        phrases = [phrase for phrase in phrases if len(phrase) > 0 and phrase != ',']
        return phrases
    
    @staticmethod
    def _phrases_to_lines(phrases):
        lines = []
        cur_level = 0

        for phrase in phrases:
            if not phrase in ['(', ')']:
                lines.append((phrase, cur_level))
            elif phrase == '(':
                cur_level += 1
            elif phrase == ')':
                cur_level -= 1
        return lines
    
    @staticmethod
    def _line_to_node(line, root=False, print_type:str = 'logical'):
        string, level = line
        if '[' in string:
            feature_importance = True
            words = [x.replace(']', '').strip() for x in string.split('[')]
            explanation_or_type = words[0]
            values = [np.abs(float(x)) for x in words[1:]]
            importance_weight = np.prod(values)
        else:
            feature_importance = False
            importance_weight = np.nan
            explanation_or_type = string
        
        if explanation_or_type in ["AND", "OR", "NOT"]:
            return Node("", level, explanation_or_type, print_type, importance_weight, feature_importance)
        else:
            return Node(explanation_or_type, level, "LEAF", print_type, importance_weight, feature_importance)

    @staticmethod
    def _text_to_tree(text, print_type):
        phrases = Explanation._text_to_phrases(text)
        lines = Explanation._phrases_to_lines(phrases)
        root = Explanation._line_to_node(lines[0], root=True, print_type=print_type)
        nodes = [Explanation._line_to_node(line, print_type=print_type) for line in lines[1:]]
        root.add_operands(nodes)
        return root
    
    @staticmethod
    def _negate(node):
        if node.type == "LEAF":
            new_node = Node(node.explanation, node.level, logic_type="NOT", print_type=node.print_type)
            new_node.convert_not_to_leaf()
            return new_node
        elif (node.type == "NOT") and (len(node.operands) < 1):
            node.parse_explanation()
            return Node(node.explanation, node.level, logic_type="LEAF", print_type=node.print_type)
        elif node.type == "NOT":
            child = node.operands[0]
            child.decrease_level()
            child = Explanation.push_negations_down(child)
            return child
        elif node.type == "AND":
            node.type = "OR"
            new_operands = [Explanation._negate(operand) for operand in node.operands]
            node.operands = new_operands
            return node
        elif node.type == "OR":
            node.type = "AND"
            new_operands = [Explanation._negate(operand) for operand in node.operands]
            node.operands = new_operands
            return node
        return Node('negation error', -1, None, None)
    
    @staticmethod
    def push_negations_down(node):
        if node.type == "LEAF":
            node.parse_explanation()
            return node
        elif (node.type == "NOT") and (len(node.operands) < 1):
            node.convert_not_to_leaf()
            return node
        elif node.type == "NOT":
            child = node.operands[0]
            new_node = Explanation._negate(child)
            new_node.decrease_level()
            return new_node
        else:
            new_node = Node(
                explanation=node.explanation, level=node.level, logic_type=node.type, print_type=node.print_type)
            new_operands = [Explanation.push_negations_down(child) for child in node.operands]
            new_node.operands = new_operands
            return new_node
               
    @staticmethod
    def _collapse_consecutive_repeated_operands(node):
        """Only works for AND and OR types of nodes, pushing negations down is needed before applying this"""
        node_type = node.type
        if not node_type in ["AND", "OR"]:
            return node
        same_type_children = list()
        different_type_children = list()
        for child in node.operands:
            if child.type == node_type:
                same_type_children.append(child)
            else:
                different_type_children.append(child)
        if len(same_type_children) < 1:
            return node
        new_children = different_type_children
        for child in same_type_children:
            child.decrease_level() # TODO replace to copy_and_decrease_level(child) to create a copy instead of changing the child
            new_children.extend(child.operands)
        new_node = Node(node.explanation, node.level, node.type, node.print_type)
        new_node.operands = new_children
        new_node = Explanation._collapse_consecutive_repeated_operands(new_node)
        return new_node

    @staticmethod
    def recursively_collapse_consecutive_repeated_operands(node):
        new_node = Explanation._collapse_consecutive_repeated_operands(node)
        new_operands = [Explanation.recursively_collapse_consecutive_repeated_operands(child) for child in new_node.operands]
        new_node.operands = new_operands
        return new_node
    
    @staticmethod
    def _negate_and_decrease_level(node):
        node.decrease_level()
        return Explanation._negate(node)
    
    @staticmethod
    def is_number(s):
        try:
            float(s) # for int, long and float
        except ValueError:
            try:
                complex(s) # for complex
            except ValueError:
                return False
        return True

    @staticmethod
    def is_integer(s):
        try:
            int(s)  # for int
        except ValueError:
            return False
        return True

    def is_integer_as_float(self, s):
        if self.is_number(s) and not self.is_integer(s):
            potential_integer = float(s)
            rounded_potential_integer = np.round(potential_integer, 0)
            if np.isclose(rounded_potential_integer, potential_integer):
                return True
        return False

    @staticmethod
    def _last_word_is_number(s):
        split = s.rsplit(' ', 1)
        if len(split) < 2:
            return (s, False)
        last_word = split[1]
        is_number = Explanation.is_number(last_word)
        return (last_word, is_number)

    @staticmethod
    def parse_explanation(x):
        """Parse a single explanation to a tuple (featuer, sign, value)"""
        last_word, is_number = Explanation._last_word_is_number(x)
        if not is_number:
            return (x, '', '')

        feature_sign = x.split(last_word)[0]

        for sep in [" is ", " was "]:
            if sep in feature_sign:
                feature_sign = feature_sign.split(sep)
        if len(feature_sign) != 2:
            print ("Complicated explanation parsing case causes error:", x)
            return (x, '', '')
        feature = feature_sign[0].strip()
        sign = feature_sign[1].strip()
        return feature, sign, last_word
    
    @staticmethod
    def recursive_explanation_parsing(node): 
        """Parse the node's explanation and all it's operands' explanations recursively"""
        node.parse_explanation()
        for child in node.operands:
            Explanation.recursive_explanation_parsing(child)

    @staticmethod
    def collapse_sample_explanation(node):
        if len(node.operands) < 1:
            return node
        
        total_operands = list()
        queue = [node]
        while queue:
            cur_node = queue.pop(0)
            for child in cur_node.operands:
                if len(child.operands) < 1:
                    new_leaf = Node(child.explanation, 1, child.type, node.print_type)
                    total_operands.append(new_leaf)
                else:
                    queue.append(child)
        final_node = Node('', 0, "AND", node.print_type)
        final_node.operands = list(set(total_operands))
        return final_node
    
    @staticmethod
    def _group_operands(node):
        feature_dict = dict()
        for operand in node.operands:
            feature = operand.feature
            if feature in feature_dict:
                feature_dict[feature].append(operand)
            else:
                feature_dict[feature] = [operand]
        return feature_dict

    @staticmethod
    def _remove_redundant_for_feature(parent_type, node_list, verbose=False):
        if not parent_type in ["AND", "OR"]:
            return node_list
        
        sign_dict = dict()
        output_list = list()
        
        for node in node_list:
            # Convert sign to its meaning (greater or less)
            sign = node.sign
            if sign in GREATER_OR_EQUAL_SIGNS:
                sign_meaning = "greater than"
                node.sign = "greater than or equal to"
            elif sign in GREATER_SIGNS:
                sign_meaning = "greater than"
                node.sign = "greater than"
            elif sign in LESS_OR_EQUAL_SIGNS:
                sign_meaning = "less than"
                node.sign = "less than or equal to"
            elif sign in LESS_SIGNS:
                sign_meaning = "less than"
                node.sign = "less than"
            else:
                output_list.append(node)
                if sign != '':
                    print("ATTENTION! Unknown sign detected: ", sign)
                continue
                
            # Adding node to sign_dict based on its sign meaning
            if sign_meaning in sign_dict:
                sign_dict[sign_meaning].append(node)
            else:
                sign_dict[sign_meaning] = [node]
                
        for sign in sign_dict:
            values = [(float(node.value), node.sign) for node in sign_dict[sign]]
            values.sort(key=lambda t: t[0], reverse=True) # descending order
            if sign in ["greater than or equal to","greater than"]:
                if parent_type == "AND":
                    final_tuple = values[0]
                elif parent_type == "OR":
                    final_tuple = values[-1]
                else:
                    print("Unknown parent type received: ", parent_type)
                    final_value = np.nan
            elif sign in ["less than or equal to", "less than"]:
                if parent_type == "AND":
                    final_tuple = values[-1]
                elif parent_type == "OR":
                    final_tuple = values[0]
                else:
                    print("Unknown parent type received: ", parent_type)
                    final_value = np.nan
            else:
                final_value = np.nan
                if sign != '':
                    print("ATTENTION! Unknown sign detected: ", sign, " sign len:", len(sign))
                output_list.extend(sign_dict[sign])
                continue
            # TODO check for contraditions like "between 10 and 1"
            final_value = final_tuple[0]
            final_sign = final_tuple[1]
            new_explanation = sign_dict[sign][0].feature + " was " + final_sign + " " + str(final_value)
            level = sign_dict[sign][0].level
            new_node = Node(new_explanation, level, "LEAF", node.print_type)
            output_list.append(new_node)
            
        if verbose:
            if len(output_list) < len(node_list):
                print("Redundancy removed from ")
                print(*node_list, sep='\n')
                print(" to ")
                print(*output_list, sep='\n')
            
        return output_list
    
    @staticmethod
    def remove_redundant(node, verbose=False):
        feature_dict = Explanation._group_operands(node)
        if len(feature_dict) < 1:
            return node
        
        total_operands = list()
        for feature in feature_dict:
            
            if feature == '':
                for deep_node in feature_dict['']:
                    total_operands.append(Explanation.remove_redundant(deep_node))
                continue
            
            simplified_operands = Explanation._remove_redundant_for_feature(node.type, \
                                                                            feature_dict[feature], verbose)
            total_operands.extend(simplified_operands)

        node.operands = total_operands
        return node
    
    @staticmethod
    def collapse_single_operand(node):
        if not node.type in ["AND", "OR"]:
            return node
        
        if len(node.operands) != 1:
            return node
        
        new_node = node.operands[0]
        new_node.decrease_level()
        return new_node
    
    @staticmethod
    def recursively_collapse_single_operands(node):
        final_operands = [Explanation.recursively_collapse_single_operands(operand) for \
                          operand in node.operands]
        node.operands = final_operands
        return Explanation.collapse_single_operand(node)
    
    @staticmethod
    def _create_between_explanation_for_AND(node_list):
        if len(node_list) != 2:
            return node_list
        # TODO check if both are leaves, not NOTs etc; check levels?
        feature = node_list[0].feature
        level = node_list[0].level
        sign1 = node_list[0].sign
        sign2 = node_list[1].sign
        value1 = node_list[0].value
        value2 = node_list[1].value
        
        if (sign1 in GREATER_SIGNS) and (sign2 in LESS_SIGNS):
            new_sign = "between"
            new_explanation = feature + " was between " + str(value1) + " and " + str(value2)
            new_node = Node(new_explanation, level, "LEAF", node_list[0].print_type)
            return [new_node]
        
        if (sign2 in GREATER_SIGNS) and (sign1 in LESS_SIGNS):
            new_sign = "between"
            new_explanation = feature + " was between " + str(value2) + " and " + str(value1)
            new_node = Node(new_explanation, level, "LEAF", node_list[0].print_type)
            return [new_node]
        
        return node_list
    
    @staticmethod
    def create_between_explanations__sample_level(node):
        """Only works for simplified tree of a sample explanation, 
        i.e. AND node with a single level of operands"""
        feature_dict = Explanation._group_operands(node)
        if len(feature_dict) < 1:
            return node
        
        total_operands = list()
        for feature in feature_dict:
            if len(feature_dict[feature]) == 1:
                total_operands.extend(feature_dict[feature])
            elif len(feature_dict[feature]) == 2:
                output = Explanation._create_between_explanation_for_AND(feature_dict[feature])
                total_operands.extend(output)
            else:
                total_operands.extend(feature_dict[feature])
        new_node = Node("", 0, "AND", node.print_type)
        new_node.operands = total_operands
        return new_node

    @staticmethod
    def create_between_explanations(node):
        feature_dict = Explanation._group_operands(node)
        if len(feature_dict) < 1:
            return node
        
        total_operands = list()
        if node.type != "AND":
            
            for feature in feature_dict:
                
                if feature == '':
                    for deep_node in feature_dict['']:
                        total_operands.append(Explanation.create_between_explanations(deep_node))
                    continue
                else:
                    total_operands.extend(feature_dict[feature])
            
            new_node = Node(node.explanation, node.level, node.type, node.print_type)
            new_node.operands = total_operands
            return new_node
            
        for feature in feature_dict:
            
            if feature == '':
                for deep_node in feature_dict['']:
                    total_operands.append(Explanation.create_between_explanations(deep_node))
                continue

            if len(feature_dict[feature]) == 1:
                total_operands.extend(feature_dict[feature])
            elif len(feature_dict[feature]) == 2:
                output = Explanation._create_between_explanation_for_AND(feature_dict[feature])
                total_operands.extend(output)
            else:
                total_operands.extend(feature_dict[feature])
        
        new_node = Node(node.explanation, node.level, node.type, node.print_type)
        new_node.operands = total_operands
        return new_node 

    @staticmethod
    def sort_node_types(node_type):
        if node_type == 'LEAF':
            return 1
        if node_type == 'NOT':
            return 2
        if node_type == 'AND':
            return 3
        return 4

    @staticmethod
    def sort_nodes(l):
        if len(l) < 1:
            return l
        expanded_list = [(node, Explanation.sort_node_types(node.type), node.explanation) for node in l]
        sorted_list = sorted(expanded_list, key=lambda x: (x[1], x[2]))
        return [t[0] for t in sorted_list]

    def format_explanation(self, node, ndigits, exclusions: list[str] = None, min_max_feature_dict: dict = None):
        new_operands = []
        if node.operands:
            for operand in node.operands:
                if operand.type == "LEAF" or ((operand.type == "NOT") and (len(operand.operands) < 1)):
                    self._round_numeric_leaves(operand, ndigits)
                    if min_max_feature_dict is not None:
                        self._add_percentile(operand, min_max_feature_dict)
                    if exclusions is not None:
                        new_operand = self._filter_excluded_leaf(operand, exclusions)
                    else:
                        new_operand = operand
                    if new_operand is not None:
                        new_operands += [new_operand]
                else:
                    new_operands += [operand]
            if new_operands:
                node.operands = new_operands
            else:
                raise AssertionError("exclusions removed all features from logic!")
        else:
            return node.operands

    def _round_numeric_leaves(self, node, ndigits):
        tokens = node.explanation.split(" ")
        tokens = [str(int(token) if self.is_integer(token)
                      else (int(np.round(float(token), 0)) if self.is_integer_as_float(token)
                            else round(float(token), ndigits=ndigits)))
                  if self.is_number(token) else token for token in tokens]
        node.explanation = " ".join(tokens)

    @staticmethod
    def _filter_excluded_leaf(node, exclusions: list[str]):
        for exclusion in exclusions:
            print(node.explanation, exclusion)
            if node.explanation.find(exclusion) > -1:
                return None
        return node

    @staticmethod
    def _add_percentile(node, min_max_feature_dict: dict):
        def _suffix(ptile):
            if ptile == '1':
                suffix = 'st'
            elif ptile == '2':
                suffix = 'nd'
            elif ptile == '3':
                suffix = 'rd'
            else:
                suffix = 'th'
            return suffix

        if node.type == 'LEAF':
            for k in min_max_feature_dict.keys():
                if node.explanation.find(k) > -1:
                    if node.explanation.find("between") > -1:
                        exp_parse = node.explanation.split(" between ")
                        exp_parse1 = [exp_parse[0]]
                        exp_parse2 = exp_parse[-1].split(" ")
                        num_1 = exp_parse2[0]
                        num_2 = exp_parse2[2]
                        ptile_1 = str(int((float(num_1) - min_max_feature_dict[k]['min'])
                                   / (min_max_feature_dict[k]['max'] - min_max_feature_dict[k]['min']) * 100))
                        ptile_2 = str(int((float(num_2) - min_max_feature_dict[k]['min'])
                                   / (min_max_feature_dict[k]['max'] - min_max_feature_dict[k]['min']) * 100))
                        node.explanation = " ".join(exp_parse1 + ["between"] + [num_1] + ["and"] + [num_2]
                                                    + [f"({ptile_1 + _suffix(ptile_1)} "
                                                       f"and {ptile_2 + _suffix(ptile_2)} percentiles)"])
                    else:
                        exp_parse = node.explanation.split(" ")
                        num = exp_parse[-1]
                        ptile = str(int((float(num) - min_max_feature_dict[k]['min'])
                                 / (min_max_feature_dict[k]['max'] - min_max_feature_dict[k]['min']) * 100))
                        node.explanation = " ".join(exp_parse[:-1] + [num] + [f"({ptile + _suffix(ptile)} percentile)"])
    

__all__ = [Node, Explanation]


import os
import joblib
import inspect
import logging
from copy import deepcopy
from typing import Tuple, Union, List
from types import FunctionType

import torch
import numpy as np
import numpy.typing as npt
import pandas as pd
from torch.utils.data import DataLoader
from sklearn.metrics import roc_auc_score, average_precision_score

from nrn.utils import tensor_agg


def _format(x):
    return x.replace("_label", "")


@joblib.delayed
def _compute_metric_func(func, predictions, col, tpn):
    p = predictions.dropna(subset=[tpn])
    t = p[tpn].values
    if not any(t == -1) and np.unique(p[tpn].values).shape[0] > 1:
        return func(p[tpn].values, p[f'{col}_{_format(tpn)}'].values)
    else:
        return np.nan


class ReasoningNetworkModel(object):

    def __init__(self):
        """
        Initialize a ReasoningNetworkModel
        """
        self.best_state = {}
        self.target_names = None
        self.rn = None

        self.USE_CUDA = torch.cuda.is_available()
        self.USE_MPS = torch.backends.mps.is_available()
        self.logger = logging.getLogger(self.__class__.__name__)

    def predict(self, dl: DataLoader) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Predict for data given by features

        Args:
            dl (DataLoader): data loader for predictions

        Returns:
            (DataFrame, DataFrame): data frame of predictions, data frame of targets
        """
        all_predictions = []
        all_targets = []
        all_indexes = []
        with torch.no_grad():
            for batch in dl:
                # [BATCH_SIZE, N_FEATURES, 2]
                features = batch['features']
                if 'target' in batch:
                    # [BATCH_SIZE, N_TARGETS]
                    target = batch['target']
                    if target.ndim > 2:
                        target.squeeze()
                    all_targets += [deepcopy(target)]

                if self.USE_CUDA:
                    features = features.cuda()
                elif self.USE_MPS:
                    features = features.to('mps')

                # [BATCH_SIZE, N_TARGETS]
                all_predictions += [self.rn(features)]
                # [BATCH_SIZE]
                all_indexes += [deepcopy(batch['sample_idx'])]

            all_predictions = torch.cat(all_predictions, dim=0).cpu().detach()
            all_indexes = torch.cat(all_indexes, dim=0)
            all_targets = torch.cat(all_targets, dim=0)

        all_predictions = all_predictions.numpy()
        all_indexes = all_indexes.numpy()
        all_targets = all_targets.numpy()

        predictions = pd.DataFrame(
            data=all_predictions,
            index=all_indexes,
            columns=[f"probs_{x.replace('_label', '')}" for x in self.target_names]
        )
        del all_predictions

        all_targets = pd.DataFrame(
            data=all_targets,
            index=all_indexes,
            columns=self.target_names
        )

        return predictions, all_targets

    def evaluate(
            self,
            predictions: pd.DataFrame,
            labels: pd.DataFrame,
            decision_boundary: float = 0.5,
            metrics: List[str] = [],
            output_metric: FunctionType = None,
            class_independent: bool = False,
            multi_class: bool = False
    ) -> Union[float, npt.NDArray]:
        """
        Evaluate predictions from features and labels

        Args:
            predictions (pd.DataFrame): data frame of predictions
            labels (pd.DataFrame): data frame of labels to evaluate against
            decision_boundary (float): (0, 1) decision boundary to use for classification
            metrics (List[str]): list of metrics to use.
                Options:  `AUC`, 'Target mAP', 'Sample mAP', `Base Rate`, `Accuracy`
            output_metric (FunctionType): A function taking 'true values', 'predicted values'
            class_independent (bool): If True, return a list of performances by class
            multi_class (bool): If True, treat as a multi-class problem, else treat as multi-label (or binary).

        Returns:
            Union[DataFrame, Int]: predictions or evaluation score
        """
        if output_metric is None:
            output_metric = roc_auc_score

        def _compute_metric(func, preds, column, metric=None, user=False):
            if not user:
                out = joblib.Parallel(n_jobs=max(1, os.cpu_count() - 1), verbose=0)(
                    _compute_metric_func(
                        func=func,
                        predictions=predictions,
                        col=column,
                        tpn=tn)
                    for tn in self.target_names)
            else:
                target_names = [x for x in self.target_names if not any(preds[x] == -1)]
                preds = preds[
                        [x for x in target_names] +
                        [f'{column}_{_format(x)}' for x in target_names]
                    ]
                out = preds.apply(lambda y: func(y[:len(target_names)], y[len(target_names):]), axis=1)

            if metric is not None:
                return metric(out)
            return np.array([round(x, 3) for x in out])

        predictions = labels.join(predictions)
        if multi_class:
            predictions_max = predictions[[f'probs_{_format(x)}' for x in self.target_names]].max(axis=1)
        else:
            predictions_max = None
        for tpn in self.target_names:
            if multi_class:
                predictions[f'predictions_{_format(tpn)}'] = predictions[f'probs_{_format(tpn)}'] == predictions_max
            else:
                predictions[f'predictions_{_format(tpn)}'] = predictions[f'probs_{_format(tpn)}'] > decision_boundary
            predictions[f'outcome_{_format(tpn)}'] = list(
                map(lambda x, y: x == y, predictions[f'predictions_{_format(tpn)}'], predictions[tpn]))

        # compute user map if multi-label
        if len(self.target_names) > 1 and 'Sample mAP' in metrics:
            mean_ap = _compute_metric(average_precision_score, predictions, "probs", np.nanmean, True)
            user_map = f'Sample mAP: {mean_ap}'
        else:
            user_map = ''

        if 'AUC' in metrics:
            avg_auc = f"Avg. AUC: {_compute_metric(roc_auc_score, predictions, 'probs', np.nanmean)}"
            med_auc = f"Med. AUC: {_compute_metric(roc_auc_score, predictions, 'probs', np.nanmedian)}"
        else:
            avg_auc = ''
            med_auc = ''

        if 'Base Rate' in metrics:
            avg_br = f"Avg. Base Rate: {predictions[self.target_names].mean(axis=1, skipna=True).mean()}"
        else:
            avg_br = ''

        if 'Accuracy' in metrics:
            acc = predictions[[f'outcome_{_format(tpn)}' for tpn in self.target_names]].mean(axis=1, skipna=True).mean()
            avg_acc = f"Avg. Accuracy: {acc}"
        else:
            avg_acc = ''

        if 'Target mAP' in metrics:
            target_map = f"Target mAP: {_compute_metric(average_precision_score, predictions, 'probs', np.nanmean)}"
        else:
            target_map = ''

        if metrics:
            self.logger.info(f"""
                    Predictions Size: {(predictions.shape[0], len(self.target_names))}
                    {avg_br}
                    {avg_acc}
                    {avg_auc}
                    {med_auc}
                    {target_map}
                    {user_map}
                    """)

        if (('y_true' in inspect.signature(output_metric).parameters)
            and ('y_pred' in inspect.signature(output_metric).parameters)
            and ('labels' in inspect.signature(output_metric).parameters)):
            col = 'predictions'
        else:
            col = 'probs'

        if not class_independent:
            return float(_compute_metric(output_metric, predictions, col, np.nanmean))
        return _compute_metric(output_metric, predictions, col, None)


__all__ = [ReasoningNetworkModel]



from copy import deepcopy
from typing import List, Tuple

import pandas as pd

import torch
from torch.utils.data import DataLoader

from .brn import BaseBanditNRNModel
from xgboost import XGBRegressor


class BoostedBanditNRNModel(BaseBanditNRNModel):

    def __init__(
            self,
            target_names: List[str],
            feature_names: List[str],
            input_size: int,
            output_size: int,
            layer_sizes: List[int],
            n_selected_features_input: int,
            n_selected_features_internal: int,
            n_selected_features_output: int,
            perform_prune_quantile: float,
            ucb_scale: float,
            normal_form: str = 'dnf',
            delta: float = 2.0,
            prune_strategy: str = 'class',
            bootstrap: bool = True,
            swa: bool = False,
            add_negations: bool = False,
            weight_init: float = 0.2,
            policy_init: torch.Tensor = None,
            xgb_max_depth: int = 4,
            xgb_n_estimators: int = 200,
            xgb_min_child_weight: float = 1.0,
            xgb_subsample: float = 0.7,
            xgb_learning_rate: float = 0.001,
            xgb_colsample_bylevel: float = 0.7,
            xgb_colsample_bytree: float = 0.7,
            xgb_gamma: float = 1,
            xgb_reg_lambda: float = 2.0,
            xgb_reg_alpha: float = 2.0,
            logits: bool = False
    ):
        """
        Initialize a Bandit Reinforced Reasoning Network model.

        Example:
            model = BaseBanditRRNModel(
                target_names=['class1', 'class2'],
                feature_names=['feature1', 'feature2', 'feature3'],
                input_size=3,
                output_size=2,
                layer_sizes=[3, 3]
                n_selected_features_input=2,
                n_selected_features_internal=2,
                n_selected_features_output=1,
                ucb_scale=1.96,
                perform_prune_quantile=0.7,
                normal_form='dnf',
                prune_strategy='class',
                bootstrap=False,
                swa=False
            )

        Args:
            target_names (list): A list of the target names.
            feature_names (list): A list of feature names.
            input_size (int): number of features from input.
            output_size (int): number of outputs.
            layer_sizes (list): A list containing the number of output logics for each layer.
            n_selected_features_input (int): The number of features to include in each logic in the input layer.
            n_selected_features_internal (int): The number of logics to include in each logic in the internal layers.
            n_selected_features_output (int): The number of logics to include in each logic in the output layer.
            perform_prune_quantile (float): The quantile to use for pruning randomized rn.
            ucb_scale (float): The scale of the confidence interval in the multi-armed bandit policy.
                               c = 1.96 is a 95% confidence interval.
            normal_form (str): 'dnf' for disjunctive normal form network; 'cnf' for conjunctive normal form network.
            delta (float): higher values increase diversity of logic generation away from existing logics.
            prune_strategy (str): Either 'class' or 'logic'.  Determines which pruning strategy to use.
            bootstrap (bool): Use boostrap samples to evaluate each atomic logic in logic prune strategy.
            swa (bool): Use stochastic weight averaging
            add_negations (bool): add negations of logic.
            weight_init (float): Upper bound of uniform weight initialization.  Lower bound is negated value.
            xgb_max_depth (int): Max depth for XGBoost boosting model.
            xgb_n_estimators (int): Number of estimators for XGBoost boosting model.
            xgb_min_child_weight (float): Minimum child weight for XGBoost boosting model.
            xgb_subsample (float): Subsample percentage for XGBoost boosting model.
            xgb_learning_rate (float): Learning rate for XGBoost boosting model.
            xgb_colsample_bylevel (float): Column subsample percent for XGBoost boosting model.
            xgb_colsample_bytree (float): Tree subsample percent for XGBoost boosting model.
            xgb_gamma (float): Gamma parameter for XGBoost boosting model.
            xgb_reg_lambda (float): Lambda regularization parameter for XGBoost boosting model.
            xgb_reg_alpha (float): Alpha regularization parameter for XGBoost boosting model.
        """
        super(BoostedBanditNRNModel, self).__init__(
            target_names=target_names,
            feature_names=feature_names,
            input_size=input_size,
            output_size=output_size,
            layer_sizes=layer_sizes,
            n_selected_features_input=n_selected_features_input,
            n_selected_features_internal=n_selected_features_internal,
            n_selected_features_output=n_selected_features_output,
            perform_prune_quantile=perform_prune_quantile,
            ucb_scale=ucb_scale,
            normal_form=normal_form,
            delta=delta,
            prune_strategy=prune_strategy,
            bootstrap=bootstrap,
            swa=swa,
            add_negations=add_negations,
            weight_init=weight_init,
            policy_init=policy_init,
            logits=logits
        )
        self.xgb_is_fitted = False
        self.initialize_booster(
            xgb_max_depth=xgb_max_depth,
            xgb_n_estimators=xgb_n_estimators,
            xgb_min_child_weight=xgb_min_child_weight,
            xgb_subsample=xgb_subsample,
            xgb_learning_rate=xgb_learning_rate,
            xgb_colsample_bylevel=xgb_colsample_bylevel,
            xgb_colsample_bytree=xgb_colsample_bytree,
            xgb_gamma=xgb_gamma,
            xgb_reg_lambda=xgb_reg_lambda,
            xgb_reg_alpha=xgb_reg_alpha,
        )

    def predict(self, dl: DataLoader) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Predict for data given by features

        Args:
            dl (DataLoader): data loader for predictions

        Returns:
            (DataFrame, DataFrame): data frame of predictions, data frame of targets
        """
        all_predictions = []
        all_targets = []
        all_indexes = []
        with torch.no_grad():
            for batch in dl:
                # [BATCH_SIZE, N_FEATURES, 2]
                features = batch['features']
                if 'target' in batch:
                    # [BATCH_SIZE, N_TARGETS]
                    target = batch['target']
                    if target.ndim > 2:
                        target.squeeze()
                    all_targets += [deepcopy(target)]

                if self.USE_CUDA:
                    features = features.cuda()
                elif self.USE_MPS:
                    features = features.to('mps')

                # [BATCH_SIZE, N_TARGETS]
                rn_predictions = self.rn(features)
                if self.xgb_is_fitted:
                    xgb_predictions = self.xgb.predict(features.cpu())
                    xgb_predictions = torch.tensor(xgb_predictions).to(rn_predictions.device)
                    if xgb_predictions.ndim < 2:
                        xgb_predictions = xgb_predictions.unsqueeze(1)
                    rn_predictions = rn_predictions + xgb_predictions
                all_predictions += [rn_predictions]
                # [BATCH_SIZE]
                all_indexes += [deepcopy(batch['sample_idx'])]

            all_predictions = torch.cat(all_predictions, dim=0).cpu().detach()
            all_indexes = torch.cat(all_indexes, dim=0)
            all_targets = torch.cat(all_targets, dim=0)

        all_predictions = all_predictions.numpy()
        all_indexes = all_indexes.numpy()
        all_targets = all_targets.numpy()

        predictions = pd.DataFrame(
            data=all_predictions,
            index=all_indexes,
            columns=[f"probs_{x.replace('_label', '')}" for x in self.target_names]
        )
        del all_predictions

        all_targets = pd.DataFrame(
            data=all_targets,
            index=all_indexes,
            columns=self.target_names
        )

        return predictions, all_targets

    def initialize_booster(
            self,
            xgb_max_depth: int = 4,
            xgb_n_estimators: int = 200,
            xgb_min_child_weight: float = 1.0,
            xgb_subsample: float = 0.7,
            xgb_learning_rate: float = 0.001,
            xgb_colsample_bylevel: float = 0.7,
            xgb_colsample_bytree: float = 0.7,
            xgb_gamma: float = 1,
            xgb_reg_lambda: float = 2.0,
            xgb_reg_alpha: float = 2.0
    ):
        self.xgb_max_depth = xgb_max_depth
        self.xgb_n_estimators = xgb_n_estimators
        self.xgb_min_child_weight = xgb_min_child_weight
        self.xgb_subsample = xgb_subsample
        self.xgb_learning_rate = xgb_learning_rate
        self.xgb_colsample_bylevel = xgb_colsample_bylevel
        self.xgb_colsample_bytree = xgb_colsample_bytree
        self.xgb_gamma = xgb_gamma
        self.xgb_reg_lambda = xgb_reg_lambda
        self.xgb_reg_alpha = xgb_reg_alpha

        self.xgb = XGBRegressor(
            learning_rate=self.xgb_learning_rate,
            gamma=self.xgb_gamma,
            max_depth=self.xgb_max_depth,
            n_estimators=self.xgb_n_estimators,
            min_child_weight=self.xgb_min_child_weight,
            subsample=self.xgb_subsample,
            colsample_bylevel=self.xgb_colsample_bylevel,
            colsample_bytree=self.xgb_colsample_bytree,
            reg_lambda=self.xgb_reg_lambda,
            reg_alpha=self.xgb_reg_alpha,
            # tree_method="hist" if self.USE_CUDA else "auto",
            # device="cuda" if self.USE_CUDA else "cpu"
        )
        self.xgb_is_fitted = False


__all__ = [BoostedBanditNRNModel]



from collections import defaultdict
from copy import deepcopy

import torch
from torch import nn
import numpy as np
import numpy.typing as npt

from ._core import LukasiewiczCore
from .predicates import BasePredicates

from .constants import *


class BaseConcatenateBlocksLogic(LukasiewiczCore):

    def __init__(self, modules, outputs_key):
        super().__init__()
        self.modules_to_concat = nn.ModuleDict({f'concat_module_{i}': m for i, m in enumerate(modules)})
        self.operands = modules
        self.outputs_key = outputs_key
        self.out_features = sum([x.out_features for x in modules])

        # create operand map
        i = 0
        self.out_feature_operand_map = defaultdict(int)
        for j, op in enumerate(self.operands):
            for k in range(op.out_features):
                self.out_feature_operand_map[i] = j
                i += 1

    def _produce_logic_string(self, operand, explanation: list, print_type='logical', depth:int = 0) -> str:
        """
        Produce a string with logical operation.

        Args:
            explanation (list): list of explanation texts.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.

        Returns:
            str: explanation text.
        """
        assert print_type in ['logical', 'logical-natural', 'natural'], \
            "'print_type' must be one of 'logical', 'logical-natural', or 'natural'"

        # joining operations for the next level down
        if print_type == 'natural':
            if operand.logic_type == 'And':
                if depth < 5:
                    joined_explanation = and_joining_options[depth].join(explanation)
                else:
                    joined_explanation = and_joining_options[-1].join(explanation)
            elif operand.logic_type == 'Or':
                if depth < 5:
                    joined_explanation = or_joining_options[depth].join(explanation)
                else:
                    joined_explanation = or_joining_options[-1].join(explanation)
        else:
            joined_explanation = ', '.join(explanation)

        if len(explanation) == 1:
            return joined_explanation
        if operand.logic_type == 'And':
            if print_type == 'logical':
                return f'AND({joined_explanation})'
            elif print_type == 'logical-natural':
                out_str = 'ALL of the following are TRUE: \n- ' + ', \n- '.join(explanation)
                return out_str.replace("\n", "\n\t")
            elif print_type == 'natural':
                if depth < 5:
                    leading_text = and_options[depth]
                else:
                    leading_text = f"{and_options[-1]}\n\n"
                return f'{leading_text}{joined_explanation}'
        if operand.logic_type == 'Or':
            if print_type == 'logical':
                return f'OR({joined_explanation})'
            elif print_type == 'logical-natural':
                out_str = 'ANY of the following are TRUE: \n- ' + ', \n- '.join(explanation)
                return out_str.replace("\n", "\n\t")
            elif print_type == 'natural':
                if depth < 5:
                    leading_text = or_options[depth]
                else:
                    leading_text = f"{or_options[-1]}\n\n"
                return f'{leading_text}{joined_explanation}'
        if operand.logic_type == 'XOr':
            if print_type == 'logical':
                return f'XOR({joined_explanation})'
            elif print_type == 'natural':
                out_str = 'ONE of the following is TRUE: \n' + ', \n- '.join(explanation)
                return out_str.replace("\n", "\n\t")

    def _produce_negation_string_natural(self, operand, text, depth):
        if depth < 5:
            condition1 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options[depth]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options[depth]) == 0)
            condition2 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options_negated[depth]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options_negated[depth]) == 0)
            if condition1:
                if operand.logic_type == 'And':
                    out_str = text.replace(and_options[depth], and_options_negated[depth], 1)
                elif operand.logic_type == 'Or':
                    out_str = text.replace(or_options[depth], or_options_negated[depth], 1)
            elif condition2:
                if operand.logic_type == 'And':
                    out_str = text.replace(and_options_negated[depth], and_options[depth], 1)
                elif operand.logic_type == 'Or':
                    out_str = text.replace(or_options_negated[depth], or_options[depth], 1)
            else:
                # handle case where text is at the predicate level.  In this case of the predicate negations are
                # flipped
                out_str = self._flip_negated_predicates(text.strip())
        else:
            condition1 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options[-1]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options[-1]) == 0)
            condition2 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options_negated[-1]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options_negated[-1]) == 0)
            if condition1:
                if operand.logic_type == 'And':
                    out_str = text.replace(and_options[-1], and_options_negated[-1], 1)
                elif operand.logic_type == 'Or':
                    out_str = text.replace(or_options[-1], or_options_negated[-1], 1)
            elif condition2:
                if operand.logic_type == 'And':
                    out_str = text.replace(and_options_negated[-1], and_options[-1], 1)
                elif operand.logic_type == 'Or':
                    out_str = text.replace(or_options_negated[-1], or_options[-1], 1)
            else:
                # handle case where text is at the predicate level.  In this case of the predicate negations are
                # flipped
                out_str = self._flip_negated_predicates(text.strip())
        out_str = self._flip_negated_logics(out_str, depth)
        return out_str

    def _produce_negation_string(
            self,
            operand,
            weight: float,
            text: str,
            negate: bool = False,
            explain_type: str = 'both',
            print_type: str = 'logical',
            depth: int = 0,
    ) -> str:
        """
        Produce a negation under certain conditions.

        Args:
            weight (float): Weight for specific predicate.
            text(str): Text for specific predicate.
            negate (bool): If True, flip operation for negation.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.

        Returns:
            str: The text with negation applied.
        """
        assert print_type in ['logical', 'logical-natural', 'natural'], \
            "'print_type' must be one of 'logical' or 'natural'"

        # conditions that produce a negative
        if (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'negative'):
            if print_type == 'logical':
                return self._produce_negation_string_logical(text)
            elif print_type == 'logical-natural':
                return self._produce_negation_string_logical_natural(text)
            elif print_type == 'natural':
                return self._produce_negation_string_natural(operand, text, depth)
            else:
                raise ValueError("`print_type` must be one of `logical` or `natural`")
        elif (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and explain_type == 'positive':
            return ""
        # conditions that produce a positive
        if (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'positive'):
            return self._natural_language_negated_predicates(text)
        elif (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and explain_type == 'negative':
            return ""
        else:
            return self._natural_language_negated_predicates(text)

    def explain_sample(
            self,
            required_output_thresholds: float,
            outputs_dict: dict = None,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: npt.NDArray = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            input_features: torch.Tensor = None,
            force_negate: bool = False,
            channel: int = 0,
            global_explain: bool = False,
            print_explanation: bool = False,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Produce a sample explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            outputs_dict (dict): Dictionary of outputs from forward pass.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            global_explain (bool): If True, perform a global explanation.
            print_explanation (bool): If True, print the explanation.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic

        Returns:
            list: explanation texts.
        """
        if isinstance(parent_weights, float) or parent_weights.ndim == 0:
            parent_weights = torch.tensor([parent_weights])

        if not global_explain:
            current_outputs = outputs_dict[self.outputs_key]
            od = deepcopy(outputs_dict)
        else:
            od = None

        # select ranges that need to be processed
        out_feature_range = torch.arange(self.out_features)
        parent_weights_mask = self._produce_weights_mask(parent_weights, quantile, threshold)

        if parent_mask is not None:
            out_feature_range = out_feature_range[parent_mask]
        if isinstance(out_feature_range, np.int64):
            out_feature_range = torch.tensor([out_feature_range])

        # process ranges
        explanation = []
        for i, (out_feature, parent_weight_mask) in enumerate(zip(out_feature_range, parent_weights_mask)):
            if parent_weight_mask:  # if the logic is above the quantile threshold then produce the explanation

                operand = self.operands[self.out_feature_operand_map[int(out_feature)]]
                if hasattr(operand, 'tau') and hasattr(operand, 'var_emb_dim'):
                    # variational
                    mask = operand.sample_mask()
                    weights_to_use = operand.weights * mask
                elif hasattr(operand, 'tau'):
                    # attn
                    if not isinstance(operand.operands, BasePredicates):
                        current_inputs = outputs_dict[operand.operands.outputs_key]
                    else:
                        current_inputs = input_features.to(operand.weights.device)
                        current_inputs = current_inputs.unsqueeze(0)
                    weights_to_use = operand.produce_explanation_weights(current_inputs)
                else:
                    weights_to_use = operand.weights
                mask_to_use = operand.mask

                if global_explain:
                    oo = self._compute_required_inputs(
                        parent_weights=parent_weights,
                        required_output_threshold=required_output_thresholds,
                        parent_logic_type=parent_logic_type,
                        negate=negate,
                        rounding_precision=rounding_precision
                    )
                    co = oo[i].squeeze()
                    oo = oo.squeeze()
                else:
                    co = current_outputs[0, channel, 0, out_feature].cpu()
                    oo = current_outputs[0, channel, 0, parent_mask].cpu().squeeze()

                current_outputs_mask, next_required_output_thresholds = self._produce_current_outputs_mask(
                    co,
                    oo,
                    parent_weights,
                    parent_weights[i],
                    parent_logic_type,
                    required_output_thresholds,
                    negate,
                    force_negate,
                    rounding_precision
                )

                if current_outputs_mask:

                    negation = negate if parent_weights[i] >= 0. else not negate
                    op_explain = operand.operands.explain_sample(
                        outputs_dict=od,
                        required_output_thresholds=next_required_output_thresholds,
                        quantile=quantile,
                        threshold=threshold,
                        parent_weights=weights_to_use[channel, :, 0].detach().cpu().squeeze(),
                        parent_mask=mask_to_use[channel, 0, :].detach().cpu(),
                        parent_logic_type=operand.logic_type,
                        negate=negation,
                        depth=depth + 1,
                        explain_type=explain_type,
                        print_type=print_type,
                        input_features=input_features,
                        channel=channel,
                        global_explain=global_explain,
                        print_explanation=print_explanation,
                        ignore_uninformative=ignore_uninformative,
                        rounding_precision=rounding_precision,
                        force_negate=force_negate,
                        inverse_transform=inverse_transform,
                        show_bounds=show_bounds,
                        **kwargs
                    )

                    if op_explain:
                        explanation_str = [
                            self._produce_negation_string(
                                operand,
                                parent_weights[i],
                                f'{self._produce_logic_string(operand, op_explain, print_type)}',
                                explain_type=explain_type,
                                print_type=print_type
                            )]
                        explanation += explanation_str

                        if print_explanation:
                            print(f"Logic at depth {depth}: {explanation_str}"
                                  f"\nweights: {weights_to_use[channel, :, 0].detach().cpu().squeeze()}"
                                  f"\noutput: {co if parent_weights[i] >= 0 else 1. - co}"
                                  f"\nrequired_threshold: {next_required_output_thresholds}\n")

        explanation = np.unique([x for x in explanation if
                                 x != "" and x.replace("\t", "").replace("\n", "") != 'NOT()']).tolist()

        if explanation:
            return explanation
        return [""]

    def explain(
            self,
            required_output_thresholds: float,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: npt.NDArray = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Produce a global explanation.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform

        Returns:
            list: explanation texts.
        """
        return self.explain_sample(
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            force_negate=force_negate,
            channel=channel,
            global_explain=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )

    def print_sample(
            self,
            outputs_dict: dict,
            required_output_thresholds: float,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: npt.NDArray = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            input_features: torch.tensor = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Print a sample explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            outputs_dict (dict): Dictionary of outputs from forward pass.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            input_features (torch.Tensor): tensor of input features.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic
        """
        return self.explain_sample(
            outputs_dict=outputs_dict,
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            input_features=input_features,
            force_negate=force_negate,
            channel=channel,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )

    def print(
            self,
            required_output_thresholds: float,
            quantile: float = 0.5,
            threshold: float = None,
            parent_weights: npt.NDArray = None,
            parent_mask: torch.Tensor = None,
            negate: bool = False,
            depth: int = 0,
            explain_type: str = 'both',
            print_type: str = 'logical',
            parent_logic_type: str = None,
            force_negate: bool = False,
            channel: int = 0,
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
            inverse_transform=None,
            show_bounds: bool = True,
            **kwargs
    ) -> list:
        """
        Produce a global explanation view of the model.

        Args:
            required_output_thresholds (float): threshold below which inner logic will be masked and excluded.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile
            parent_weights (np.array): Array of parent weights.
            parent_mask (np.array): Array of parent mask.
            negate (bool): If True, flip operation for negation.
            depth (int): Depth of the current logic in the network.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.
            parent_logic_type (str): one of 'Or', 'And'.
            force_negate (bool): If True, extract the negation of logic.
            channel (int): channel to perform traversal over.
            ignore_uninformative (bool): If True, ignore uninformative logic that has threshold at 0 or 1.
            rounding_precision (int): precision for rounding percentiles during explanation.
            inverse_transform (sklearn transform): inverse of in the inputs transform
            show_bounds (bool): include numeric boundary used in logic
        """
        return self.explain_sample(
            required_output_thresholds=required_output_thresholds,
            quantile=quantile,
            threshold=threshold,
            parent_weights=parent_weights,
            parent_mask=parent_mask,
            negate=negate,
            depth=depth,
            explain_type=explain_type,
            print_type=print_type,
            parent_logic_type=parent_logic_type,
            force_negate=force_negate,
            channel=channel,
            global_explain=True,
            print_explanation=True,
            ignore_uninformative=ignore_uninformative,
            rounding_precision=rounding_precision,
            inverse_transform=inverse_transform,
            show_bounds=show_bounds,
            **kwargs
        )



from .rn import ReasoningNetworkModel
from .pruningrn import PruningReasoningNetworkModel
from .brn import BaseBanditNRNModel
from .boosted_brn import BoostedBanditNRNModel


__all__ = [ReasoningNetworkModel, PruningReasoningNetworkModel, BaseBanditNRNModel, BoostedBanditNRNModel]



import numpy as np
import pandas as pd
from sklearn.metrics import roc_auc_score

from nrn.utils.mic import compute_mic_matrix

import torch
from torch import nn
from torch import optim
from torch.utils.data import DataLoader

from nrn.models import BanditNRNClassifier
from nrn.utils.trainers import BanditNRNTrainer

from ..datasets.simple_dataset import SimpleDataset
from ..base.base_estimator import BaseSKLogicEstimator


# class RNRNClassifier(BaseSKLogicEstimator):
class NRNClassifier(BaseSKLogicEstimator):

    def __init__(
            self,
            target_names: list = None,
            feature_names: list = None,
            layer_sizes: list = [8, 8],
            n_selected_features_input: int = 4,
            n_selected_features_internal: int = 4,
            n_selected_features_output: int = 4,
            perform_prune_quantile: float = 0.5,
            ucb_scale: float = 1.5,
            prune_strategy: str = 'class',
            delta: float = 2.0,
            bootstrap: bool = False,
            swa: bool = False,
            add_negations: bool = False,
            normal_form: str = 'cnf',
            weight_init: float = 0.2,
            logits: bool = True,
            binarization: bool = True,
            tree_num: int = 10,
            tree_depth: int = 5,
            tree_feature_selection: float = 0.5,
            thresh_round: int = 3,
            loss_func=nn.BCEWithLogitsLoss,
            learning_rate: float = 0.1,
            weight_decay: float = 0.001,
            t_0: int = 3,
            t_mult: int = 2,
            epochs: int = 200,
            batch_size: int = 32,
            holdout_pct: float = 0.2,
            early_stopping_plateau_count: int = 20,
            perform_prune_plateau_count: int = 3,
            increase_prune_plateau_count: int = 10,
            increase_prune_plateau_count_plateau_count: int = 10,
            lookahead_steps: int = 0,
            lookahead_steps_size: float = 0.0,
            evaluation_metric=roc_auc_score,
            multi_class: bool = False,
            pin_memory: bool = False,
            persistent_workers: bool = False,
            num_workers: bool = False
    ):
        """
        RNRN Scikit-Learn compatible classifier.

    `   Args:
            target_names List[str]: List of names for targets
            feature_names List[str]: List of desired feature names
            layer_sizes List[int]: List of layer sizes corresponding to width of each layer
            n_selected_features_input (int): Number of input features selected in each input logic
            n_selected_features_internal (int): Number of hidden features selected in each hidden logic
            n_selected_features_output (int): Number of hidden features selected for each output logic
            perform_prune_quantile (float): Quantile of model to prune during pruning phase
            ucb_scale (float): UCB scale to use for multi-armed bandit
            prune_strategy (str): Prune strategy, one of 'class', 'logic', 'class_logic'
            delta (float): Factor used to reduce likelihood of sampling un-pruned logic during growth phase
            bootstrap (bool): If using logic prune strategy, evaluate with bootstrap sampling
            swa (bool): If true, use swa weight averaging
            add_negations (bool): If true, add negated logic on initialization
            normal_form (str): Conjunctive normal form or disjunctive normal form structure, 'cnf' or 'dnf'
            weight_init (float): Weigh initialization magnitude
            logits (bool): If true, output are logits, otherwise unscaled.
            binarization (bool): if true, use feature binarization from trees
            tree_num (int): number of trees for feature binarization from trees
            tree_depth (int): depth of trees for feature binarization from trees
            tree_feature_selection (float): feature selection rate for feature binarization from trees
            thresh_round (int): rounding threshold in decimal places for feature binarzation from trees
            loss_func (Loss): PyTorch loss
            learning_rate (float): learning rate for AdamW optimizer
            weight_decay (float): weight decay for AdamW optimizer
            t_0 (int): T_0 for CosineAnnealingWithWarmRestarts scheduler
            t_mult (int): T_mult for CosineAnnealingWithWarmRestarts scheduler
            epochs (int): epochs for training procedure
            batch_size (int): batch size for data loading
            holdout_pct (float): percentage of training data used as holdout for early stopping
            early_stopping_plateau_count (int): number of epochs without improvement for early stopping
            perform_prune_plateau_count (int): Perform pruning phase after plateau count epochs without improvement
            increase_prune_plateau_count (int): Increase the perform_prune_plateau_count by this number if plateaued
            increase_prune_plateau_count_plateau_count (int): Increase perform_prune_plateau_count when reached .
            lookahead_steps (int): number of steps for lookahead optimization. If zero, lookahead not used.
            lookahead_steps_size (float): step size for lookahead optimization.  If lookahead_steps is zero not used.
            evaluation_metric (metric): Scikit-Learn metric function
            multi_class (bool): If true, treat as multi-class problem, otherwise multi-label or binary
            pin_memory (bool): pin_memory for data loaders.
            persistent_workers (bool): persistent workers for data loaders.
            num_workers (int): number of workers for data loaders.
        """
        # super(RNRNClassifier, self).__init__(
        super(NRNClassifier, self).__init__(
            binarization=binarization,
            tree_num=tree_num,
            tree_depth=tree_depth,
            tree_feature_selection=tree_feature_selection,
            thresh_round=thresh_round,
            learning_rate=learning_rate,
            weight_decay=weight_decay,
            t_0=t_0,
            t_mult=t_mult,
            epochs=epochs,
            batch_size=batch_size,
            holdout_pct=holdout_pct,
            early_stopping_plateau_count=early_stopping_plateau_count,
            lookahead_steps=lookahead_steps,
            lookahead_steps_size=lookahead_steps_size,
            pin_memory=pin_memory,
            persistent_workers=persistent_workers,
            num_workers=num_workers
        )
        # handle empty target names
        if target_names:
            self.target_names = target_names
        else:
            self.target_names = None

        # handle empty feature names
        if feature_names:
            self.feature_names = feature_names
        else:
            self.feature_names = None

        # add hyper-parameters
        self.layer_sizes = layer_sizes
        self.n_selected_features_input = n_selected_features_input
        self.n_selected_features_internal = n_selected_features_internal
        self.n_selected_features_output = n_selected_features_output
        self.perform_prune_quantile = perform_prune_quantile
        self.ucb_scale = ucb_scale
        self.prune_strategy = prune_strategy
        self.delta = delta
        self.bootstrap = bootstrap
        self.swa = swa
        self.add_negations = add_negations
        self.normal_form = normal_form
        self.weight_init = weight_init
        self.logits = logits
        self.loss_func = loss_func
        self.perform_prune_plateau_count = perform_prune_plateau_count
        self.increase_prune_plateau_count = increase_prune_plateau_count
        self.increase_prune_plateau_count_plateau_count = increase_prune_plateau_count_plateau_count
        self.evaluation_metric = evaluation_metric
        self.multi_class = multi_class

    def fit(self, X: pd.DataFrame, y: pd.DataFrame) -> None:
        """
        Fit the model

        Args:
            X (pd.DataFrame): Dataframe of features
            y (pd.DataFrame): Dataframe of targets

        Returns:
            None
        """
        if not isinstance(X, pd.DataFrame):
            X = self._handle_non_dataframe_features(X)
        X = X.copy()

        if not isinstance(y, pd.DataFrame):
            y = self._handle_non_dataframe_targets(y)
        y = y.copy()

        X = self._handle_empty_feature_names(X)

        # create textual class names if they are not given by user
        if self.target_names is None:
            self.target_names = [f"Class {i}" for i in range(y.shape[1])]
            y.columns = self.target_names

        # ecode data in necessary format
        X = self._fit_transform_encode_data(X)
        if self.binarization:
            X = self._fit_transform_binarize_features(X, y)
        feature_names = X.columns

        # pytorch data
        dataset = SimpleDataset(X.values, y.values)
        train_dl, train_holdout_dl = self._generate_training_data_loaders(dataset)

        # initial bandit policy
        assert len(feature_names) == X.shape[1], f"feat names: {len(feature_names)}; x: {X.shape[1]}"
        mic_c_policy, _ = compute_mic_matrix(X, y, alpha=.45, c=6)
        mic_c_policy = torch.tensor(mic_c_policy)

        if X.shape[1] < self.n_selected_features_input:
            Warning(
                "The number of features is less than 'n_selected_features_input'.  Using number of features instead.")
            self.n_selected_features_input = X.shape[1]

        # init model
        self.model = BanditNRNClassifier(
            target_names=self.target_names,
            feature_names=list(feature_names),
            input_size=len(feature_names),
            output_size=len(self.target_names),
            layer_sizes=self.layer_sizes,
            n_selected_features_input=self.n_selected_features_input,
            n_selected_features_internal=self.n_selected_features_internal,
            n_selected_features_output=self.n_selected_features_output,
            perform_prune_quantile=self.perform_prune_quantile,
            ucb_scale=self.ucb_scale,
            prune_strategy=self.prune_strategy,
            normal_form=self.normal_form,
            delta=self.delta,
            bootstrap=self.bootstrap,
            swa=self.swa,
            add_negations=self.add_negations,
            weight_init=self.weight_init,
            logits=self.logits,
            policy_init=mic_c_policy
        )

        optimizer = optim.AdamW(self.model.rn.parameters(), lr=self.learning_rate, weight_decay=self.weight_decay)
        scheduler = optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=self.t_0, T_mult=self.t_mult)
        trainer = BanditNRNTrainer(
            model=self.model,
            loss_func=self.loss_func(),
            optimizer=optimizer,
            scheduler=scheduler,
            epochs=self.epochs,
            accumulation_steps=1,
            l1_lambda=0.,
            early_stopping_plateau_count=self.early_stopping_plateau_count,
            perform_prune_plateau_count=self.perform_prune_plateau_count,
            increase_prune_plateau_count=self.increase_prune_plateau_count,
            increase_prune_plateau_count_plateau_count=self.increase_prune_plateau_count_plateau_count,
            lookahead_steps=self.lookahead_steps,
            lookahead_steps_size=self.lookahead_steps_size,
            augment=None,
            augment_alpha=0.
        )

        # train model
        # The trainer defaults to optimizing the validation roc_auc_score.
        # To optimize against a different metric pass the sklearn metric to the 'evaluation_metric' parameter
        trainer.train(train_dl, train_holdout_dl, evaluation_metric=self.evaluation_metric,
                      multi_class=self.multi_class)
        trainer.set_best_state()

    def predict(self, X: pd.DataFrame, decision_boundary: float = 0.5) -> pd.DataFrame:
        """
        Predict classes with fitted model on new data.

        Args:
            X (pd.DataFrame): Untransformed input features

        Returns:
            pd.DataFrame: Predictions from fitted model
        """
        assert self.model is not None, "must fit before prediction"

        if not isinstance(X, pd.DataFrame):
            X = self._handle_non_dataframe_features(X)
        X = X.copy()

        X = self._handle_empty_feature_names(X)

        X = self._encode_prediction_data(X)
        dataset = SimpleDataset(X.values, np.ones(shape=(X.shape[0], self.model.output_size)))
        prediction_dl = DataLoader(
            dataset, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, persistent_workers=self.persistent_workers, num_workers=self.num_workers
            # very important to optimize these settings in production
        )

        predictions, _ = self.model.predict(prediction_dl)
        # class_predictions = (predictions > predictions.median()).astype(int)
        class_predictions = (predictions > decision_boundary).astype(int)

        class_predictions.rename(columns=lambda x: x.replace("probs_", ""), inplace=True)

        return class_predictions

    def predict_proba(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Predict probabilities with fitted model on new data.

        Args:
            X (pd.DataFrame): Untransformed input features

        Returns:
            pd.DataFrame: Predictions from fitted model
        """
        assert self.model is not None, "must fit before prediction"

        if not isinstance(X, pd.DataFrame):
             X = self._handle_non_dataframe_features(X)
        X = X.copy()

        X = self._handle_empty_feature_names(X)

        X = self._encode_prediction_data(X)
        dataset = SimpleDataset(X.values, np.ones(shape=(X.shape[0], self.model.output_size)))
        prediction_dl = DataLoader(
            dataset, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, persistent_workers=self.persistent_workers, num_workers=self.num_workers
            # very important to optimize these settings in production
        )

        predictions, _ = self.model.predict_proba(prediction_dl)

        predictions.rename(columns=lambda x: x.replace("probs_", ""), inplace=True)

        return predictions

    def score(self, X: pd.DataFrame, y: pd.DataFrame) -> float:
        """
        Score a model with new data using the evaluation metric

        Args:
            X (pd.DataFrame): Untransformed input features
            y (pd.DataFrame): target input

        Returns:
            float: score from evaluation metric
        """
        assert self.model is not None, "must fit before scoring"

        if not isinstance(X, pd.DataFrame):
            X = self._handle_non_dataframe_features(X)
        X = X.copy()

        if not isinstance(y, pd.DataFrame):
            y = self._handle_non_dataframe_targets(y)
        y = y.copy()

        X = self._handle_empty_feature_names(X)

        X = self._encode_prediction_data(X)

        dataset = SimpleDataset(X.values, y.values)

        prediction_dl = DataLoader(
            dataset, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, persistent_workers=self.persistent_workers, num_workers=self.num_workers
            # very important to optimize these settings in production
        )

        predictions, _ = self.model.predict_proba(prediction_dl)

        return self.evaluation_metric(y.values, predictions.values)

    def _get_instance_params(self):
        """
        Get parameters for this model type.

        Returns:
            dict: parameters and values
        """
        return {
            'layer_sizes': self.layer_sizes,
            'n_selected_features_input': self.n_selected_features_input,
            'n_selected_features_internal': self.n_selected_features_internal,
            'n_selected_features_output': self.n_selected_features_output,
            'perform_prune_quantile': self.perform_prune_quantile,
            'ucb_scale': self.ucb_scale,
            'prune_strategy': self.prune_strategy,
            'delta': self.delta,
            'bootstrap': self.bootstrap,
            'swa': self.swa,
            'add_negations': self.add_negations,
            'weight_init': self.weight_init,
            'logits': self.logits,
            'loss_func': self.loss_func,
            'perform_prune_plateau_count': self.perform_prune_plateau_count,
            'increase_prune_plateau_count': self.increase_prune_plateau_count,
            'increase_prune_plateau_count_plateau_count': self.increase_prune_plateau_count_plateau_count,
            'evaluation_metric': self.evaluation_metric,
            'multi_class': self.multi_class,
        }
    
    @staticmethod
    def _beautify_output(out):
        if not isinstance(out, str):
            return out
        if not out.startswith("0: "):
            return out
        return out[3:]

    def explain_sample(
            self,
            X: pd.DataFrame,
            sample_index: int = 0,
            quantile: float = 1.0,
            decision_boundary: float = 0.5
    ) -> str:
        """
        Generate a sample explanation

        Args:
            X (pd.DataFrame): DataFrame of input features.
            sample_index (int): Index of sample to explain.
            quantile (float): Percent of model to explain
            decision_boundary (float): Decision boundary for positive class

        Returns:
            str: explanation for selected sample
        """
        assert not self.multi_class, "explanation not currently supported for multi-class"
        assert self.model is not None, "must fit before explaining"

        if not isinstance(X, pd.DataFrame):
            X = self._handle_non_dataframe_features(X)
        X = X.copy()

        X = self._handle_empty_feature_names(X)

        X = self._encode_prediction_data(X)

        self.min_max_features_dict = {
            col: {'min': self.numeric_features.iloc[:, i].min(), 'max': self.numeric_features.iloc[:, i].max()}
            for i, col in enumerate(self.numeric_features.columns)
        }

        dataset = SimpleDataset(X.values, np.ones(shape=(X.shape[0], 1)))

        out = self.model.explain_samples(
            dataset[sample_index]['features'].unsqueeze(0),
            quantile=quantile,
            target_names=self.target_names,
            explain_type='both',
            sample_explanation_prefix="The prediction is",
            print_type='logical',
            ignore_uninformative=True,
            rounding_precision=3,
            show_bounds=not self.binarization,
            decision_boundary=decision_boundary,
            simplify=True,
            exclusions=None,
            min_max_feature_dict=self.min_max_features_dict,
            feature_importances=False
        )
        return NRNClassifier._beautify_output(out)



import numpy as np
import torch.nn as nn


class BaseReasoningNetworkMixin(nn.Module):

    def __init__(self):
        super(BaseReasoningNetworkMixin, self).__init__()
        self.model = None
        self.output_layer = None
        self.root_layer = None
        self._feature_importances = dict()

    def set_modules(
            self,
            model,
            root_layer=None
    ):
        if self.USE_DATA_PARALLEL:
            self.model = model.module.model
            self.output_layer = model.module.output_layer
            if root_layer is None:
                self.root_layer = model.module.output_layer
            else:
                self.root_layer = root_layer
        else:
            self.model = model.model
            self.output_layer = model.output_layer
            if root_layer is None:
                self.root_layer = model.output_layer
            else:
                self.root_layer = root_layer

    @staticmethod
    def _build_search_space(
            quantile: float = 0.5,
            explain_type: str = 'both',
            ignore_uninformative: bool = False,
            rounding_precision: int = 3,
    ):
        search_params = []

        # 1. rounding_precision
        # 2. ignore_uninformative
        # 3. explain_type
        # 3. ignore_uninformative and explain_type
        # 4. ignore_uninformative and rounding_precision
        # 5. ignore_uninformative and explain_type and rounding_precision
        # 6. quantile
        # 7. quantile and rounding_precision
        # 8. quantile and ignore_uninformative
        # 9. quantile and explain_type
        # 9. quantile and ignore_uninformative and explain_type
        # 9. quantile and rounding_precision and ignore_uninformative
        # 9. quantile and rounding_precision and ignore_uninformative and explain_type

        # search over rounding precisions
        for i in range(6, 15):
            search_params += [
                {'quantile': quantile,
                 'explain_type': explain_type,
                 'ignore_uninformative': ignore_uninformative,
                 'rounding_precision': i,
                 }
            ]

        # search over ignore_uninformative
        search_params += [
            {'quantile': quantile,
             'explain_type': explain_type,
             'ignore_uninformative': False,
             'rounding_precision': rounding_precision,
             }
        ]

        # search over explain_type
        search_params += [
            {'quantile': quantile,
             'explain_type': 'both',
             'ignore_uninformative': ignore_uninformative,
             'rounding_precision': rounding_precision,
             }
        ]

        # search over ignore_uninformative and explain_type
        search_params += [
            {'quantile': quantile,
             'explain_type': 'both',
             'ignore_uninformative': False,
             'rounding_precision': rounding_precision,
             }
        ]

        # search over ignore_uninformative and rounding precisions
        for i in range(6, 15):
            search_params += [
                {'quantile': quantile,
                 'explain_type': explain_type,
                 'ignore_uninformative': False,
                 'rounding_precision': i,
                 }
            ]

        # search over ignore_uninformative and explain_type and rounding precisions
        for i in range(6, 15):
            search_params += [
                {'quantile': quantile,
                 'explain_type': 'both',
                 'ignore_uninformative': False,
                 'rounding_precision': i,
                 }
            ]

        # search over quantile
        for i in np.arange(quantile, 1.0, 0.05):
            search_params += [
                {'quantile': i,
                 'explain_type': explain_type,
                 'ignore_uninformative': ignore_uninformative,
                 'rounding_precision': rounding_precision,
                 }
            ]

        # search over quantile and rounding_precision
        for i in np.arange(quantile, 1.0, 0.05):
            for j in range(6, 15):
                search_params += [
                    {'quantile': i,
                     'explain_type': explain_type,
                     'ignore_uninformative': ignore_uninformative,
                     'rounding_precision': j,
                     }
                ]

        # search over quantile and ignore_uninformative
        for i in np.arange(quantile, 1.0, 0.05):
            search_params += [
                {'quantile': i,
                 'explain_type': explain_type,
                 'ignore_uninformative': False,
                 'rounding_precision': rounding_precision,
                 }
            ]

        # search over quantile and explain_type
        for i in np.arange(quantile, 1.0, 0.05):
            search_params += [
                {'quantile': i,
                 'explain_type': 'both',
                 'ignore_uninformative': ignore_uninformative,
                 'rounding_precision': rounding_precision,
                 }
            ]

        # search over quantile and ignore_uninformative and explain_type
        for i in np.arange(quantile, 1.0, 0.05):
            search_params += [
                {'quantile': i,
                 'explain_type': 'both',
                 'ignore_uninformative': False,
                 'rounding_precision': rounding_precision,
                 }
            ]

        # search over quantile and rounding_precision and ignore_uninformative
        for i in np.arange(quantile, 1.0, 0.05):
            for j in range(6, 16):
                search_params += [
                    {'quantile': i,
                     'explain_type': explain_type,
                     'ignore_uninformative': False,
                     'rounding_precision': j,
                     }
                ]

        # search over quantile and rounding_precision and explain_type
        for i in np.arange(quantile, 1.0, 0.05):
            for j in range(6, 15):
                search_params += [
                    {'quantile': i,
                     'explain_type': 'both',
                     'ignore_uninformative': ignore_uninformative,
                     'rounding_precision': j,
                     }
                ]

        # search over quantile and rounding_precision and explain_type and ignore_uninformative
        for i in np.arange(quantile, 1.0, 0.05):
            for j in range(6, 15):
                search_params += [
                    {'quantile': i,
                     'explain_type': 'both',
                     'ignore_uninformative': False,
                     'rounding_precision': j,
                     }
                ]

        return search_params



import re
from nrn.utils.explanations.simplification import Explanation


def remove_duplicate_words(input1: str):
    # Regex to matching repeated words
    regex = r'\b(\w+)(?:\W+\1\b)+'

    return re.sub(regex, r'\1', input1, flags=re.IGNORECASE)


def remove_character_combo(input1: str, char1: str, char2: str, replacement_char: str):
    # retplace the string ";," with ";"

    return input1.replace(f"{char1}{char2}", f"{replacement_char}")


def get_outputs(name, outputs):
    def hook(model, input, output):
        if isinstance(output, tuple):
            outputs[name] = output[0].detach()
        else:
            outputs[name] = output.detach()

    return hook


def register_hooks(model, outputs, mode='explanation'):
    for x in model.named_children():
        x[1].register_forward_hook(get_outputs(x[0], outputs))
        register_hooks(x[1], outputs)


def simplification(
        explanation_str,
        print_type,
        simplify,
        sample_level=False,
        verbose=False,
        ndigits: int = 3,
        exclusions: list[str] = None,
        min_max_feature_dict: dict = None,
        feature_importances: bool = False
):
    if verbose:
        print("\n_____\nInput string:\n\n", explanation_str)

    if simplify:
        assert print_type in ['logical', 'logical-natural'], \
            "print_type must be 'logical' or 'logical-natural' if simplify is True"
        
    if feature_importances:
        exp = Explanation(explanation_str, print_type)
        exp.root.sort_operands()
        if verbose:
            print("\n_____\nExplanation tree for feature importances (not simplified):\n", exp.root.tree_to_string())
        return exp.root

    if print_type == 'natural':
        return explanation_str
    
    exp = Explanation(explanation_str, print_type)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nExplanation tree:\n", exp.root.tree_to_string())

    if not simplify:
        exp.root.sort_operands()
        # return exp.root.tree_to_string()
        return exp.root
    
    exp.root = Explanation.push_negations_down(exp.root)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nNegations pushed down:\n\n", exp.root.tree_to_string())

    exp.root = Explanation.recursively_collapse_consecutive_repeated_operands(exp.root)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nConsecutive operands collapsed:\n", exp.root.tree_to_string())

    Explanation.recursive_explanation_parsing(exp.root)
    exp.root = Explanation.remove_redundant(exp.root)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nRedundant features removed:\n", exp.root.tree_to_string())

    exp.root = Explanation.recursively_collapse_single_operands(exp.root)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nSingle operands collapsed:\n", exp.root.tree_to_string())

    Explanation.recursive_explanation_parsing(exp.root)
    exp.root = Explanation.remove_redundant(exp.root)
    if verbose:
        exp.root.sort_operands()
        print("\n_____\nRedundant features removed:\n", exp.root.tree_to_string())

    if not sample_level:
        Explanation.recursive_explanation_parsing(exp.root) # Adding feature, sign, value to each node
        exp.simple_root = Explanation.create_between_explanations(exp.root)
        exp.simple_root.sort_operands()
        if verbose:
            print("\n_____\nBetween explanations created:\n", exp.simple_root.tree_to_string())
        # return exp.simple_root.tree_to_string() 
        return exp.simple_root 
    
    # Now we know that simplify is True, print_type is 'logical' and sample_level is True
    # Sample level simplification -- for sample level explanations only!
    exp.sample_root = Explanation.collapse_sample_explanation(exp.root)
    if verbose:
        exp.sample_root.sort_operands()
        print("\n_____\nThe tree collapsed to sample level:\n", exp.sample_root.tree_to_string())
    
    Explanation.recursive_explanation_parsing(exp.sample_root)
    exp.sample_root = Explanation.remove_redundant(exp.sample_root, verbose=verbose)
    if verbose:
        exp.sample_root.sort_operands()
        print("\n_____\nSample level redundant features removed:\n", exp.sample_root.tree_to_string())

    Explanation.recursive_explanation_parsing(exp.sample_root) # To add feature, sign, value to each node
    exp.sample_root = Explanation.create_between_explanations__sample_level(exp.sample_root)
    exp.sample_root.sort_operands()
    if verbose:
        print("\n_____\nSample level between explanations created:\n", exp.sample_root.tree_to_string())

    Explanation.format_explanation(exp, exp.sample_root, ndigits, exclusions, min_max_feature_dict)
    if verbose:
        print("\n_____\nSample level formatted explanations created:\n", exp.sample_root.tree_to_string())

    exp.sample_root.sort_operands()
    # return exp.sample_root.tree_to_string()
    return exp.sample_root




import os
import glob

from typing import Union, Tuple

import numpy as np

import torch
import torch.nn as nn

from .constants import *


class LukasiewiczCore(nn.Module):

    def __init__(self):
        super(LukasiewiczCore, self).__init__()

    def __produce_joined_explanation(self, explanation, print_type, depth):
        # joining operations for the next level down
        if print_type == 'logical':
            joined_explanation = ', '.join(explanation)
        elif print_type == 'logical-natural':
            joined_explanation = ', \n- '.join(explanation)
        elif print_type == 'natural':
            if self.logic_type == 'And':
                if depth < 5:
                    joined_explanation = and_joining_options[depth].join(explanation)
                else:
                    joined_explanation = and_joining_options[-1].join(explanation)
            elif self.logic_type == 'Or' or self.logic_type == 'XOr':
                if depth < 5:
                    joined_explanation = or_joining_options[depth].join(explanation)
                else:
                    joined_explanation = or_joining_options[-1].join(explanation)
            else:
                raise ValueError("logic_type must be 'And' or 'Or', 'XOr'.")
        else:
            raise ValueError("'print_type' must be one of 'logical' or 'natural'")
        return joined_explanation

    def __produce_logic_string_and(self, joined_explanation, print_type, depth):
        if print_type == 'logical':
            return f'AND({joined_explanation})'
        elif print_type == 'logical-natural':
            out_str = 'ALL the following are TRUE: \n- ' + joined_explanation
            return out_str.replace("\n", "\n\t")
        elif print_type == 'natural':
            if depth < 5:
                leading_text = and_options[depth]
            else:
                leading_text = f"{and_options[-1]}\n\n"
            return f'{leading_text}{joined_explanation}'

    def __produce_logic_string_or(self, joined_explanation, print_type, depth):
        if print_type == 'logical':
            return f'OR({joined_explanation})'
        elif print_type == 'logical-natural':
            out_str = 'ANY of the following are TRUE: \n- ' + joined_explanation
            return out_str.replace("\n", "\n\t")
        elif print_type == 'natural':
            if depth < 5:
                leading_text = or_options[depth]
            else:
                leading_text = f"{or_options[-1]}\n\n"
            return f'{leading_text}{joined_explanation}'

    def __produce_logic_string_xor(self, joined_explanation, print_type, depth):
        if print_type == 'logical':
            return f'XOR({joined_explanation})'
        elif print_type == 'logical-natural':
            out_str = 'ONE of the following is TRUE: \n- ' + joined_explanation
            return out_str.replace("\n", "\n\t")

    def _produce_logic_string(
            self, explanation: list,
            print_type: str = 'logical',
            depth: int = 0
    ) -> str:
        """
        Produce a string with logical operation.

        Args:
            explanation (list): list of explanation texts.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.

        Returns:
            str: explanation text.
        """
        assert print_type in ['logical', 'logical-natural', 'natural'], \
            "'print_type' must be one of 'logical' or 'natural'"

        joined_explanation = self.__produce_joined_explanation(explanation, print_type, depth)

        # add leading text
        if len(explanation) == 1:
            return joined_explanation
        if self.logic_type == 'And':
            return self.__produce_logic_string_and(joined_explanation, print_type, depth)
        if self.logic_type == 'Or':
            return self.__produce_logic_string_or(joined_explanation, print_type, depth)
        if self.logic_type == 'XOr':
            return self.__produce_logic_string_xor(joined_explanation, print_type, depth)

    @staticmethod
    def _natural_language_negated_predicates(text):
        out_str = text.split(",")
        out_str_conv = []
        for pred_str in out_str:
            if pred_str.find("NOT") > -1:
                if pred_str.find("and NOT ") > -1:
                    out_str_conv += [pred_str.replace("and NOT ", "and it was NOT true ")]
                elif pred_str.find("or NOT ") > -1:
                    out_str_conv += [pred_str.replace("or NOT ", "or it was NOT true ")]
                elif pred_str[:4] == 'NOT ' and pred_str[:17] != "NOT the following":
                    out_str_conv += [f"it was NOT true {pred_str.replace('NOT ', '', 1)}"]
                else:
                    out_str_conv += [pred_str]
            else:
                out_str_conv += [pred_str]
        out_str = ','.join(out_str_conv)
        return out_str

    @staticmethod
    def _flip_negated_logics(text, depth):
        if depth > 5:
            raise Warning("Natural language explanations with a network depth greater than 6 levels and negations"
                          "are not guaranteed to produce correct results")

        and_joining_str = and_joining_options[depth].replace(".", "").strip()
        or_joining_str = or_joining_options[depth].replace(".", "").strip()
        and_joining_str_negated = and_joining_options_negated[depth].replace(".", "").strip()
        or_joining_str_negated = or_joining_options_negated[depth].replace(".", "").strip()

        out_str = text.split(".")
        out_str_conv = []
        for sub_str in out_str:
            if sub_str.find(and_joining_str) > -1:
                out_str_conv += [sub_str.replace(and_joining_str, and_joining_str_negated)]
            elif sub_str.find(or_joining_str) > -1:
                out_str_conv += [sub_str.replace(or_joining_str, or_joining_str_negated)]
            else:
                out_str_conv += [sub_str]
        out_str = ".".join(out_str_conv)
        return out_str

    @ staticmethod
    def _flip_negated_predicates(text):
        # handle case where text is at the predicate level.  In this case of the predicate negations are
        # flipped
        out_str = text.split(",")
        out_str_conv = []
        for pred_str in out_str:
            if pred_str.find("NOT") == -1:
                if pred_str.find("- ") > -1:
                    out_str_conv += [pred_str.replace("- ", "- NOT ")]
                elif pred_str.find(" and ") == 0:
                    out_str_conv += [pred_str.replace("and ", "or it was NOT true ")]
                elif pred_str.find(" or ") == 0:
                    out_str_conv += [pred_str.replace("or ", "and it was NOT true ")]
                else:
                    out_str_conv += ["it was NOT true " + pred_str]
            else:
                if pred_str.find("- NOT") > -1:
                    out_str_conv += [pred_str.replace("- NOT ", "- ")]
                elif pred_str.find("and it was NOT true") > -1:
                    out_str_conv += [pred_str.replace("and it was NOT true ", "or ")]
                elif pred_str.find("and NOT") > -1:
                    out_str_conv += [pred_str.replace("and NOT ", "or ")]
                elif pred_str.find("or it was NOT true") > -1:
                    out_str_conv += [pred_str.replace("or it was NOT true ", "and ")]
                elif pred_str.find("or NOT") > -1:
                    out_str_conv += [pred_str.replace("or NOT ", "and ")]
                elif pred_str.find("it was NOT true") > -1:
                    out_str_conv += [pred_str.replace("it was NOT true ", "")]
                elif pred_str.find("NOT ") > -1:
                    out_str_conv += [pred_str.replace("NOT ", "")]
        out_str = ','.join(out_str_conv)
        return out_str

    @staticmethod
    def _produce_negation_string_logical(text):
        # if (not text.replace('\n', '').replace('\t- ', '')[:8] == 'NOT(AND('
        if (not text.replace('\n', '').replace('\t- ', '')[:8] in ['NOT(AND(', 'NOT(AND[']
                # and not text.replace('\n', '').replace('\t- ', '')[:7] == 'NOT(OR('):
                and not text.replace('\n', '').replace('\t- ', '')[:7] in ['NOT(OR(', 'NOT(OR[']):
            out_str = f"NOT({text})"
        else:
            out_str = text.replace("NOT(", '', 1)[:-1]
        return out_str

    def _produce_negation_string_logical_natural(self, text):
        if not text.replace('\n', '').replace('\t- ', '')[:17] == 'NOT the following':
            out_str = f"NOT the following: \n- {text}"  # original
        elif text.replace('\n', '').replace('\t- ', '').find(
                "NOT the following:") == 0:
            out_str = text.replace(
                "NOT the following:", "", 1).replace('\n\t- ', '').strip()
        else:
            out_str = self._flip_negated_predicates(text)
        return out_str.replace("\n", "\n\t")

    def _produce_negation_string_natural(self, text, depth):
        if depth < 5:
            condition1 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options[depth]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options[depth]) == 0)
            condition2 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options_negated[depth]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options_negated[depth]) == 0)
            if condition1:
                if self.logic_type == 'And':
                    out_str = text.replace(and_options[depth], and_options_negated[depth], 1)
                elif self.logic_type == 'Or':
                    out_str = text.replace(or_options[depth], or_options_negated[depth], 1)
            elif condition2:
                if self.logic_type == 'And':
                    out_str = text.replace(and_options_negated[depth], and_options[depth], 1)
                elif self.logic_type == 'Or':
                    out_str = text.replace(or_options_negated[depth], or_options[depth], 1)
            else:
                # handle case where text is at the predicate level.  In this case of the predicate negations are
                # flipped
                out_str = self._flip_negated_predicates(text.strip())
        else:
            condition1 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options[-1]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options[-1]) == 0)
            condition2 = (text.lstrip('\n').lstrip("\t").lstrip().find(and_options_negated[-1]) == 0
                          or text.lstrip('\n').lstrip("\t").lstrip().find(or_options_negated[-1]) == 0)
            if condition1:
                if self.logic_type == 'And':
                    out_str = text.replace(and_options[-1], and_options_negated[-1], 1)
                elif self.logic_type == 'Or':
                    out_str = text.replace(or_options[-1], or_options_negated[-1], 1)
            elif condition2:
                if self.logic_type == 'And':
                    out_str = text.replace(and_options_negated[-1], and_options[-1], 1)
                elif self.logic_type == 'Or':
                    out_str = text.replace(or_options_negated[-1], or_options[-1], 1)
            else:
                # handle case where text is at the predicate level.  In this case of the predicate negations are
                # flipped
                out_str = self._flip_negated_predicates(text.strip())
        out_str = self._flip_negated_logics(out_str, depth)
        return out_str

    def _produce_negation_string(
            self,
            weight: float,
            text: str,
            negate: bool = False,
            explain_type: str = 'both',
            print_type: str = 'logical',
            depth: int = 0
    ) -> str:
        """
        Produce a negation under certain conditions.

        Args:
            weight (float): Weight for specific predicate.
            text(str): Text for specific predicate.
            negate (bool): If True, flip operation for negation.
            explain_type (str): One of `both`, `positive`, `negative`. Control types of logic to produce.
            print_type (str): one of 'logical' or 'natural'.  If logical, prints logic, otherwise, prints natural text.

        Returns:
            str: The text with negation applied.
        """
        assert print_type in ['logical', 'logical-natural', 'natural'], \
            "'print_type' must be one of 'logical' or 'natural'"

        # conditions that produce a negative
        if (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'negative'):
            if print_type == 'logical':
                return self._produce_negation_string_logical(text)
            elif print_type == 'logical-natural':
                return self._produce_negation_string_logical_natural(text)
            elif print_type == 'natural':
                return self._produce_negation_string_natural(text, depth)
            else:
                raise ValueError("`print_type` must be one of `logical` or `natural`")
        elif (((weight < 0 and not negate) or (weight >= 0 and negate)) and text != "") \
                and explain_type == 'positive':
            return ""
        # conditions that produce a positive
        if (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and (explain_type == 'both' or explain_type == 'positive'):
            return self._natural_language_negated_predicates(text)
        elif (((weight < 0 and negate) or (weight >= 0 and not negate)) and text != "") \
                and explain_type == 'negative':
            return ""
        else:
            return self._natural_language_negated_predicates(text)

    @staticmethod
    def _produce_weights_mask(weights: torch.Tensor, quantile: float = None, threshold: float = None) -> torch.Tensor:
        """
        Produce a mask based on weights.

        Args:
            weights (np.array): Array of weights.
            quantile (float): Quantile of logic to produce.
            threshold (float): Weight threshold to use.  If not None, then overrides quantile

        Returns:
            np.array: Boolean mask on weights
        """
        zero_weights = weights == 0
        if isinstance(zero_weights, torch.BoolType):
            zero_weights = torch.tensor(zero_weights)
        if torch.all(zero_weights):
            return ~zero_weights
        if threshold is None:
            threshold = torch.quantile(torch.abs(weights[~zero_weights]), 1 - quantile)
        return torch.tensor(torch.abs(weights) >= threshold) & torch.tensor(torch.abs(weights) > 0)

    def _produce_current_outputs_mask(
            self,
            current_outputs: torch.Tensor,
            input_features: float,
            all_parent_weights,
            parent_weights: torch.Tensor,
            parent_logic_type: str,
            required_output_threshold: torch.Tensor,
            negate: bool = False,
            force_negate: bool = False,
            rounding_precision: int = 3,
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        Produce a mask based on the current outputs.

        Args:
            current_outputs (Union[npt.NDArray, float, int]): the outputs from the current logic.
            parent_weights (Union[npt.NDArray, float, int]): the parent weights for the current logic.
            parent_logic_type (str): one of 'Or', 'And'.
            required_output_threshold (float): the required output threshold, below which the mask is False.
            negate (bool): If True, flip operation for negation.

        Returns:
            Tuple[npt.NDArray, npt.NDArray]: mask, thresholds
        """
        # if the logic is negated then the result of the current logic can't produce an output
        # greater than 1 - required thresholds
        ROUND_PRECISION = 32

        if negate:
            required_output_threshold = 1 - required_output_threshold
        required_output_threshold = torch.round(required_output_threshold, decimals=ROUND_PRECISION)

        aws = all_parent_weights.greater_equal(torch.tensor(0.0)).float()
        input_features_negated = aws * input_features + (1.0 - aws) * (1.0 - input_features)

        ws = parent_weights.greater_equal(torch.tensor(0.0)).float()
        current_outputs_negated = ws * current_outputs + (1.0 - ws) * (1.0 - current_outputs)

        if parent_logic_type == 'Or':
            # compute other outputs
            other_outputs = (input_features_negated @ torch.abs(all_parent_weights)
                             - current_outputs_negated * torch.abs(parent_weights))
            other_outputs = torch.round(other_outputs, decimals=ROUND_PRECISION)

            # force to zero
            if torch.logical_and(other_outputs >= -1e-7, other_outputs <= 1e-7):
                other_outputs = torch.tensor(0.0)

            # compute thresholds
            thresholds = torch.clip(
                (required_output_threshold - other_outputs) / torch.abs(parent_weights), 0.0, 1.0)
            thresholds = torch.round(thresholds, decimals=ROUND_PRECISION)

        elif parent_logic_type == 'And':
            # compute other outputs
            other_outputs = ((1 - input_features_negated) @ torch.abs(all_parent_weights)
                             - (1 - current_outputs_negated) * torch.abs(parent_weights))
            other_outputs = torch.round(other_outputs, decimals=ROUND_PRECISION)

            # force to zero
            if torch.logical_and(other_outputs >= -1e-7, other_outputs <= 1e-7):
                other_outputs = torch.tensor(0.0)

            # compute thresholds
            thresholds = torch.clip(
                1.0 - (1.0 - required_output_threshold - other_outputs) / np.abs(parent_weights),
                0.0, 1.0)
            thresholds = torch.round(thresholds, decimals=ROUND_PRECISION)
        elif parent_logic_type == 'XOr':
            # compute other outputs
            other_outputs = ((1 - input_features_negated) @ np.abs(all_parent_weights)
                             - (1 - current_outputs_negated) * np.abs(parent_weights))
            other_outputs = torch.round(other_outputs, decimals=ROUND_PRECISION)

            # force to zero
            if torch.logical_and(other_outputs >= -1e-7, other_outputs <= 1e-7):
                other_outputs = torch.tensor(0.0)

            # compute thresholds
            thresholds = torch.clip(
                1.0 - (1.0 - required_output_threshold + other_outputs) / torch.abs(parent_weights),
                0.0, 1.0)
            thresholds = torch.round(thresholds, decimals=ROUND_PRECISION)
        else:
            raise ValueError("parent logic type must be 'Or', 'And', 'XOr'.")

        # When force negating we don't want to change the threshold, but we want to have values below the threshold
        # instead of above it
        negate = negate if not force_negate else not negate

        # TODO: There seems to be some level of numerical instability that causes these values to not work out
        #  exactly.  This seems to be fixed after converting code to use PyTorch

        # def trunc(values, decs=0):
        #     return np.trunc(values * 10 ** decs) / (10 ** decs)
        #
        # def round_down(values, decs=0):
        #     return np.floor(values * 10 ** decs) / (10 ** decs)
        #
        # if negate:
        #     current_outputs_negated = round_down(current_outputs_negated, rounding_precision)
        #     thresholds = np.round(thresholds, rounding_precision)
        # else:
        #     current_outputs_negated = np.round(current_outputs_negated, rounding_precision)
        #     thresholds = round_down(thresholds, rounding_precision)

        # # TODO: Remove once testing for explanations is completed
        # if negate:
        #     m = current_outputs_negated.lt(thresholds)
        # else:
        #     if force_negate:
        #         m = current_outputs_negated.gt(thresholds)
        #     else:
        #         m = current_outputs_negated.ge(thresholds) | torch.allclose(current_outputs_negated, thresholds)
        # print(type(self), "CO", current_outputs, "CON", current_outputs_negated, "T", thresholds,
        #       "RT", required_output_threshold, "W", parent_weights, "APW", all_parent_weights,
        #       "IF", input_features, "IFNEG", input_features_negated,
        #       "M", m, "NEGATE", negate, "OO", other_outputs, "PT", parent_logic_type)

        # if the logic is negated then the result of the current logic can't produce an output
        # greater than 1 - required thresholds
        if negate:
            # return current_outputs_negated.lt(thresholds), thresholds
            return current_outputs_negated.lt(thresholds), thresholds, current_outputs_negated, other_outputs
        if force_negate:
            # if force_negate then both conditions must be strict inequalities since we are negating
            # return current_outputs_negated.gt(thresholds), thresholds
            return current_outputs_negated.gt(thresholds), thresholds, current_outputs_negated, other_outputs
        # return current_outputs_negated.ge(thresholds) | torch.allclose(current_outputs_negated, thresholds), thresholds
        return current_outputs_negated.ge(thresholds) | torch.allclose(current_outputs_negated, thresholds), thresholds, current_outputs_negated, other_outputs

    def _compute_required_inputs(
            self,
            parent_weights: torch.Tensor,
            required_output_threshold: torch.Tensor,
            parent_logic_type: str,
            negate: bool = False,
            force_negate: bool = False,
            rounding_precision: int = 3
    ) -> torch.Tensor:
        if isinstance(parent_weights, torch.Tensor):
            parent_weights = parent_weights.numpy()
        if isinstance(required_output_threshold, torch.Tensor):
            required_output_threshold = required_output_threshold.numpy()

        if negate:
            required_output_threshold = 1.0 - required_output_threshold

        # handle zero weights
        zero_weight_mask = parent_weights == 0
        non_zero_indices = np.where(~zero_weight_mask)[0]
        parent_weights = parent_weights[~zero_weight_mask]

        # adjust threshold for logic type
        if parent_logic_type == 'And':
            required_output_threshold = 1.0 - required_output_threshold
        elif parent_logic_type == 'Or':
            pass
        elif parent_logic_type == 'XOr':
            required_output_threshold = 1.0 - required_output_threshold
        else:
            raise AssertionError("`parent_logic_type` must be one of 'And', 'Or', 'XOr'.")

        # sign of weights to flip required inputs where necessary
        if force_negate:
            weights_sign = (parent_weights < 0).astype(int)
        else:
            weights_sign = (parent_weights > 0).astype(int)

        # handle parent weights that can't fully contribute
        parent_weights_too_small = np.round(np.abs(parent_weights)
                                            - required_output_threshold/parent_weights.shape[0], rounding_precision)
        parent_weights_too_small_mask = parent_weights_too_small < 0.0
        if np.any(parent_weights_too_small_mask):
            parent_weights_not_too_small_indices = np.where(~parent_weights_too_small_mask)[0]
            parent_weights_too_small_indices = np.where(parent_weights_too_small_mask)[0]
            # if none of the weights are large enough then there is nothing to distribute.
            if not np.any(~parent_weights_too_small_mask):
                parent_weights_too_small_mask = np.array([False])
            else:
                parent_weights_too_small_diff = parent_weights_too_small[parent_weights_too_small_mask].sum()
                parent_weights_too_small_diff_spread = (parent_weights_too_small_diff /
                                                        sum(~parent_weights_too_small_mask) * -1)

        # matrix of one-hot weights
        parent_weights = (parent_weights.reshape(
            1, -1).repeat(parent_weights.shape[0], axis=0)
                          * np.identity(parent_weights.shape[0]))

        # vector of required_output_thresholds / count of weights
        required_output_threshold = np.array(
            required_output_threshold).repeat(parent_weights.shape[0]) \
                                    / parent_weights.shape[0]

        # adjust required output thresholds based on weights that were too small
        if np.any(parent_weights_too_small_mask):
            required_output_threshold_pos_adjustment = np.zeros_like(required_output_threshold)
            required_output_threshold_neg_adjustment = np.zeros_like(required_output_threshold)
            np.put_along_axis(
                required_output_threshold_pos_adjustment, parent_weights_not_too_small_indices,
                parent_weights_too_small_diff_spread, axis=0)
            np.put_along_axis(
                required_output_threshold_neg_adjustment, parent_weights_too_small_indices,
                parent_weights_too_small_diff, axis=0)
            required_output_threshold = (required_output_threshold
                                         + required_output_threshold_pos_adjustment
                                         + required_output_threshold_neg_adjustment)

        # computes the input required for each to contribute equally to the required output
        required_inputs = np.linalg.solve(np.abs(parent_weights), required_output_threshold)

        # adjust required inputs for logic type
        if parent_logic_type == 'And':
            required_inputs = 1.0 - required_inputs
        elif parent_logic_type == 'Or':
            pass
        elif parent_logic_type == 'XOr':
            required_inputs = 1.0 - required_inputs
        else:
            raise AssertionError("`parent_logic_type` must be one of 'And', 'Or', 'XOr'.")

        required_inputs = np.round(
            weights_sign * required_inputs + (1.0 - weights_sign) * (1.0 - required_inputs), rounding_precision)

        # # TODO: remove this if you won't use a strictly < condition for negations
        # # ensure that inputs are strictly greater than or less than the required threshold
        # if negate:
        #     ws_adj_sub = (weights_sign.astype(bool)).astype(int)
        #     ws_adj_add = (~weights_sign.astype(bool)).astype(int)
        # else:
        #     ws_adj_sub = (~weights_sign.astype(bool)).astype(int)
        #     ws_adj_add = (weights_sign.astype(bool)).astype(int)
        # adj_sub = np.ones_like(required_inputs) / (10 ** rounding_precision) * ws_adj_sub
        # adj_add = np.ones_like(required_inputs) / (10 ** rounding_precision) * ws_adj_add
        # required_inputs -= adj_sub
        # required_inputs += adj_add
        #
        # # # TODO: Remove when testing is complete
        # # print("ADJ_ADD", adj_add, "ADJ_SUB", adj_sub, "WS_A", ws_adj_add, "WS_S", ws_adj_sub, "WS", weights_sign, "PW",
        # #       parent_weights, "RI", required_inputs, "NEG", negate)

        # handle zero weights by reconstructing array of original size
        if np.any(zero_weight_mask):
            tmp_required_inputs = np.zeros_like(zero_weight_mask.astype(float))
            np.put_along_axis(tmp_required_inputs, non_zero_indices, required_inputs, axis=0)
            required_inputs = tmp_required_inputs

        required_inputs = np.round(required_inputs, rounding_precision)

        # # TODO: Remove after testing is done
        # print("RI", required_inputs, "ROT", required_output_threshold,
        #       "PW", parent_weights.sum(axis=1), "PLT", parent_logic_type)

        return torch.tensor(np.clip(required_inputs, 0.0, 1.0)).float()

    @staticmethod
    def _generate_feature_importance(
            current_outputs_negated: torch.Tensor,
            parent_weight: torch.Tensor,
            feature_importances_type: str,
            parent_logic_type: str,
            required_output_thresholds: torch.Tensor,
            other_outputs: torch.Tensor
    ) -> float:
        """
        Produce one of the feature importance values

        Args:
            current_outputs_negated (Tensor): The negated outputs of current nodes being currently evaluated
            parent_weight (Tensor): The parent weight on the current output
            i (int): The index in parent weights corresponding to the current output being evaluated
            feature_importances_type (str): The type of feature importance value to return
            parent_logic_type (str): The logic of the parent node
            required_output_thresholds (Tensor): The required output of the node
            other_outputs (Tensor): The value of the other outputs in the currently evaluated logic

        Returns:
            float: feature importance
        """
        feature_value = current_outputs_negated.item()
        contribution_weight = np.abs(parent_weight.item())
        if feature_importances_type == 'weight':
            importance_value = contribution_weight
        elif feature_importances_type == "weight_proportion":
            importance_value = contribution_weight / parent_weight.abs().sum().item()
            # print("predicate", weight, contribution_weight, parent_weights.abs().sum().item(), importance_value)
        elif feature_importances_type == "influence":
            if parent_logic_type == "And":
                importance_value = 1. - (1. - feature_value) * contribution_weight
            elif parent_logic_type == "Or":
                importance_value = feature_value * contribution_weight
            else:
                importance_value = np.nan
        elif feature_importances_type == 'incremental_influence':
            importance_value = min(max(required_output_thresholds.item() - other_outputs.item(), 0.), 1.)
        elif feature_importances_type == 'capped_weight':
            if feature_value == 0. and parent_logic_type == "And":
                importance_value = contribution_weight
            elif feature_value == 0. and parent_logic_type == "Or":
                importance_value = 0.
            else:
                importance_value = min(contribution_weight, 1. / feature_value)
        else:
            importance_value = np.nan

        return importance_value



from .distributed import grad_agg, float_agg, tensor_agg
from .explanations import simplification, remove_duplicate_words

__all__ = [grad_agg, float_agg, tensor_agg, simplification, remove_duplicate_words]



import torch
import numpy as np
from torch.utils.data import Dataset


class SimpleDataset(Dataset):

    """
    Class for a simple dataset used in sklogic.
    """

    def __init__(
            self,
            X: np.array,
            y: np.array
    ):
        """
        Dataset suitable for SKLogic models model from torchlogic

        Args:
            X (np.array): features data scaled to [0, 1]
            y (np.array): target data of classes 0, 1
        """
        super(SimpleDataset, self).__init__()
        self.X = X.astype('float')
        self.y = y
        self.sample_idx = np.arange(X.shape[0])  # index of samples

    def __len__(self):
        return self.X.shape[0]

    def __getitem__(self, idx):
        features = torch.from_numpy(self.X[idx, :]).float()
        target = torch.from_numpy(self.y[idx, :])
        return {'features': features, 'target': target, 'sample_idx': idx}


import math
from typing import List
from typing import Optional as _Optional

import torch
from torch import nn
from torch import Tensor

import numpy as np


def val_clamp(x, _min: float = 0.0, _max: float = 1.0) -> torch.Tensor:
    """gradient-transparent clamping to clamp values between [min, max]"""
    clamp_min = (x.detach() - _min).clamp(max=0)
    clamp_max = (x.detach() - _max).clamp(min=0)
    return x - clamp_max - clamp_min


def soft_val_clamp(x):
    m = torch.nn.Sigmoid()
    return m(x)


def alpha_clamp(x: torch.Tensor, alpha: float):
    """
    Clamps values lte to alph to zero and gte to alpha to one.
    """
    assert 1.0 >= alpha >= 0.0, "alpha must be between 0 and 1."

    with torch.no_grad():
        true_mask = torch.Tensor(x > (1.0 - alpha))
        false_mask = torch.Tensor(x <= alpha)
        true_delta = (1.0 - x) * true_mask.to(torch.float)
        false_delta = -1.0 * x * false_mask.to(torch.float)
    return x + true_delta + false_delta


def minmax_scale(input_tensor: torch.Tensor, _min: float = 0, _max: float = 1) -> torch.Tensor:
    """scale a tensor between _min and _max"""
    v_min, v_max = input_tensor.min(), input_tensor.max()
    return (input_tensor - v_min) / (v_max - v_min) * (_max - _min) + _min


def get_outputs(name, outputs):
    def hook(model, input, output):
        outputs[name] = output.detach()

    return hook


def register_hooks(model, outputs):
    for x in model.named_children():
        x[1].register_forward_hook(get_outputs(x[0], outputs))
        if not isinstance(x[1], nn.ModuleList):
            register_hooks(x[1], outputs)



def adjust_alpha(alpha: float, eps: float = 1e-8) -> float:
    """
    Adjust the alpha value such that it is in (0.5, 1]

    Returns:
        Dict: relations as groundings
    """
    if alpha == 0.5:
        alpha += eps
        return alpha
    else:
        return alpha


def chunk_inputs(
        inputs: List,
        fan_in: int = 4,
        inputs_probs: List[float] = None
) -> List:
    """
    Chunk the inputs according to the selection criteria mode

    Args:
        inputs (List): List of inputs
        fan_in (int): number of inputs to group together
        inputs_probs (List[float]): list of probabilities to use for sampling inputs

    Returns:
        List: list of chunked inputs
    """
    inputs = np.random.choice(inputs, size=len(inputs), replace=False, p=inputs_probs)
    chunked_inputs = [inputs[i:i + fan_in] for i in range(0, len(inputs), fan_in)]

    if len(chunked_inputs) == 0:
        raise AssertionError("Chunking inputs resulted in list of zero length.  Try adjusting drop rates.")

    return chunked_inputs


def exu(x, weight, bias):
    """ExU hidden unit modification."""
    return torch.exp(weight) * (x - bias)


class EXU(nn.Module):
    r"""Applies a linear transformation to the incoming data: :math:`y = xA^T + b`.

    This module supports :ref:`TensorFloat32<tf32_on_ampere>`.

    On certain ROCm devices, when using float16 inputs this module will use :ref:`different precision<fp16_on_mi200>` for backward.

    Args:
        in_features: size of each input sample
        out_features: size of each output sample
        bias: If set to ``False``, the layer will not learn an additive bias.
            Default: ``True``

    Shape:
        - Input: :math:`(*, H_{in})` where :math:`*` means any number of
          dimensions including none and :math:`H_{in} = \text{in\_features}`.
        - Output: :math:`(*, H_{out})` where all but the last dimension
          are the same shape as the input and :math:`H_{out} = \text{out\_features}`.

    Attributes:
        weight: the learnable weights of the module of shape
            :math:`(\text{out\_features}, \text{in\_features})`. The values are
            initialized from :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})`, where
            :math:`k = \frac{1}{\text{in\_features}}`
        bias:   the learnable bias of the module of shape :math:`(\text{out\_features})`.
                If :attr:`bias` is ``True``, the values are initialized from
                :math:`\mathcal{U}(-\sqrt{k}, \sqrt{k})` where
                :math:`k = \frac{1}{\text{in\_features}}`

    Examples::

        >>> m = nn.Linear(20, 30)
        >>> input = torch.randn(128, 20)
        >>> output = m(input)
        >>> print(output.size())
        torch.Size([128, 30])
    """

    __constants__ = ['in_features', 'out_features']
    in_features: int
    out_features: int
    weight: Tensor

    def __init__(self, in_features: int, out_features: int, bias: bool = True,
                 device=None, dtype=None) -> None:
        factory_kwargs = {'device': device, 'dtype': dtype}
        super().__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.weight = nn.Parameter(torch.empty((1, out_features), **factory_kwargs))
        if bias:
            self.bias = nn.Parameter(torch.empty((1, out_features), **factory_kwargs))
        else:
            self.register_parameter('bias', None)
        self.reset_parameters()

    def reset_parameters(self) -> None:
        # Setting a=sqrt(5) in kaiming_uniform is the same as initializing with
        # uniform(-1/sqrt(in_features), 1/sqrt(in_features)). For details, see
        # https://github.com/pytorch/pytorch/issues/57109
        # nn.init.kaiming_uniform_(self.weight, a=math.sqrt(5))
        nn.init.trunc_normal_(self.weight, mean=4, std=0.5, a=3.0, b=5.0)
        if self.bias is not None:
            nn.init.trunc_normal_(self.bias, std=0.5, a=-1.0, b=1.0)

    @staticmethod
    def exu(input, weight, bias):
        """ExU hidden unit modification."""
        return torch.exp(weight) * (input - bias)

    def forward(self, input: Tensor) -> Tensor:
        return self.exu(input, self.weight, self.bias)

    def extra_repr(self) -> str:
        return f'in_features={self.in_features}, out_features={self.out_features}, bias={self.bias is not None}'


def _no_grad_normal_(tensor, mean, std, generator=None):
    with torch.no_grad():
        return tensor.normal_(mean, std, generator=generator)


def _no_grad_uniform_(tensor, a, b, generator=None):
    with torch.no_grad():
        return tensor.uniform_(a, b, generator=generator)


def lukasiewicz_channel_block_xavier_uniform_(
    tensor: Tensor, gain: float = 1.0, generator: _Optional[torch.Generator] = None
) -> Tensor:
    r"""Fill the input `Tensor` with values using a Xavier uniform distribution.

    The method is described in `Understanding the difficulty of training
    deep feedforward neural networks` - Glorot, X. & Bengio, Y. (2010).
    The resulting tensor will have values sampled from
    :math:`\mathcal{U}(-a, a)` where

    .. math::
        a = \text{gain} \times \sqrt{\frac{6}{\text{fan\_in} + \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        gain: an optional scaling factor
        generator: the torch Generator to sample from (default: None)

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.xavier_uniform_(w, gain=nn.init.calculate_gain('relu'))
    """
    fan_in = tensor.size(1)
    fan_out = 1

    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))
    a = math.sqrt(3.0) * std  # Calculate uniform bounds from standard deviation

    return _no_grad_uniform_(tensor, -a, a, generator)


def lukasiewicz_channel_block_xavier_normal_(
    tensor: Tensor,
    gain: float = 1.0,
    generator: _Optional[torch.Generator] = None,
) -> Tensor:
    r"""Fill the input `Tensor` with values using a Xavier normal distribution.

    The method is described in `Understanding the difficulty of training deep feedforward
    neural networks` - Glorot, X. & Bengio, Y. (2010). The resulting tensor
    will have values sampled from :math:`\mathcal{N}(0, \text{std}^2)` where

    .. math::
        \text{std} = \text{gain} \times \sqrt{\frac{2}{\text{fan\_in} + \text{fan\_out}}}

    Also known as Glorot initialization.

    Args:
        tensor: an n-dimensional `torch.Tensor`
        gain: an optional scaling factor
        generator: the torch Generator to sample from (default: None)

    Examples:
        >>> w = torch.empty(3, 5)
        >>> nn.init.xavier_normal_(w)
    """
    fan_in = tensor.size(1)
    fan_out = 1

    std = gain * math.sqrt(2.0 / float(fan_in + fan_out))

    return _no_grad_normal_(tensor, 0., std, generator)


__all__ = [soft_val_clamp, minmax_scale, adjust_alpha, chunk_inputs]



import logging
from nrn.models.base import ReasoningNetworkModel


class PruningReasoningNetworkModel(ReasoningNetworkModel):

    def __init__(self):
        """
        Initialize a PruningReasoningNetworkModel
        """
        super(PruningReasoningNetworkModel, self).__init__()
        self.best_state = {}
        self.target_names = None
        self.rn = None
        self.logger = logging.getLogger(self.__class__.__name__)

    def perform_prune(self) -> None:
        """
        Update the nodes in the PruningLogicalNetwork by pruning and growing.
        """
        raise Warning("Abstract method, must be implemented!")


__all__ = [PruningReasoningNetworkModel]



from .predicates import Predicates
from .blocks import (LukasiewiczChannelAndBlock, LukasiewiczChannelOrBlock, LukasiewiczChannelXOrBlock)
from .utils import ConcatenateBlocksLogic

__all__ = [Predicates,
           LukasiewiczChannelAndBlock,
           LukasiewiczChannelOrBlock,
           LukasiewiczChannelXOrBlock,
           ConcatenateBlocksLogic]



keyword_constraints = ['scenarios', 'requirements', 'conditions', 'situations', 'circumstances', 'context']

or_options = ["There are several scenarios that could be met for this prediction to hold true.  "
              "The first scenario that could be met is as follows.  ",
              "At least one of the following requirements are met.  The first requirement is as follows. ",
              "There are several more conditions that could be met.  The conditions are described next.  ",
              "In fact, there are additional situations that could be met.  ",
              "It must be true that at least one of the following circumstances are met.  "
              "The first circumstance is the following.  ",
              "There is additional context that could be met.  The contexts are described next.  "]
and_options = ["There are several scenarios that must be met for this prediction to hold true.  "
               "The first scenario is as follows.  ",
               "All of the following requirements are met. The first requirement is as follows.  ",
               "There are several more conditions that must be met.  The conditions are described next.  ",
               "In fact, there are additional situations that must be met.  ",
               "It must be true that the following circumstances are met.  The first circumstance is the following.  ",
               "There is additional context that must be met.  The contexts are described next.  "]

or_options_negated = ["There are several scenarios that must NOT be met for this prediction to hold true. "
                      "The first scenario that must NOT be met is as follows.  ",
                      "All of the following requirements are NOT met. The first requirement is as follows.  ",
                      "There are several more conditions that must NOT be met.  The conditions are described next.  ",
                      "In fact, there are additional situations that must NOT be met.  "
                      "The first situation is as follows. ",
                      "It must be true that all of the following circumstances are NOT met.  "
                      "The first circumstance is the following.  ",
                      "There is additional context that must NOT be met.  The contexts are described next.  "]
and_options_negated = ["There are several scenarios, at least one of which must NOT hold,"
                       " for this prediction to hold true. "
                       "The first scenario is as follows.  ",
                       "It must be true that at least one of the "
                       "following requirements are NOT met.  The first requirement is as follows. ",
                       "There are several more conditions, at least one of which must NOT be met.  "
                       "The conditions are described next.  ",
                       "In fact, there are additional situations, at least one of which must NOT be met.  ",
                       "It must be true that at least one of the following circumstances are NOT met.  "
                       "The first circumstance is the following.  ",
                       "There is additional context, at least one of which must NOT be met.  "
                       "The contexts are described next.  "]

and_joining_options = [".\n\nThe next scenario that must be met is as follows.  ",
                       ".  An additional requirement that must be met is the following.  ",
                       ".  As well as the following conditions.  ",
                       ".\n\nThe next situation that must be met is as follows.  ",
                       ".  An additional circumstance that must be met is the following.  ",
                       ".  As well as the following context.  "]
or_joining_options = [".\n\nThe next scenario that could be met is as follows.  ",
                      ".  An additional requirement that could be met is the following.  ",
                      ".  Or the following conditions.  ",
                      ".\n\nThe next situation that could be met is as follows.  ",
                      ".  An additional circumstance that could be met is the following.  ",
                      ".  Or the following context.  "]


and_joining_options_negated = [".\n\nThe next scenario that could NOT be met is as follows.  ",
                               ".  An additional requirement that could NOT be met is the following.  ",
                               ".  Or the following conditions are NOT met.  ",
                               ".\n\nThe next situation that coule NOT be met is as follows.  ",
                               ".  An additional circumstance that could NOT be be met is the following.  ",
                               ".  Or the following context is NOT met.  "]
or_joining_options_negated = [".\n\nThe next scenario that must NOT be met is as follows.  ",
                              ".  An additional requirement that must NOT be met is the following.  ",
                              ".  The following conditions are also NOT met.  ",
                              ".\n\nThe next situation that must NOT be met is as follows.  ",
                              ".  An additional circumstance that must NOT be met is the following.  ",
                              ".  And the following context is NOT met.  "]


import numpy as np

import torch
from torch.utils.data import DataLoader

from nrn.models.base import BaseBanditNRNModel
from .banditnrntrainer import BanditNRNTrainer


class BoostedBanditNRNTrainer(BanditNRNTrainer):

    def __init__(
            self,
            model: BaseBanditNRNModel,
            loss_func,
            optimizer,
            scheduler=None,
            epochs: int = 100,
            accumulation_steps: int = 1,
            objective: str = 'maximize',
            l1_lambda: float = 1e-5,
            lookahead_steps: int = 5,
            lookahead_steps_size: float = 0.7,
            augment: str = None,
            augment_alpha: float = 0.03,
            class_independent: bool = False,
            early_stopping_plateau_count: int = 20,
            perform_prune_plateau_count: int = 10,
            increase_prune_plateau_count: int = 20,
            increase_prune_plateau_count_plateau_count: int = 20,
            boosting_epochs: int = 1,
            partial_fit: bool = False
    ):
        """
        Train a BanditRRN model.

        Example:
            trainer = BanditRRNTrainer(
                model=model,
                loss_func=nn.CrossEntropyLoss(),
                optimizer=optim.Adam,
                epochs=100,
                learning_rate=learning_rate,
                accumulation_steps=1,
                l1_lambda=l1_lambda,
                weight_decay=weight_decay,
                early_stopping_plateau_count=early_stopping_plateau_count,
                perform_prune_plateau_count=perform_prune_plateau_count,
                explain_percentile=0.9,
                verbose=0,
                distributed=False
            )

            trainer.train(train_dl, val_dl)

        Args:
            model (BanditRRNModel): the model to be trained.
                Format:
                    This model must implement the following methods:
                        predict(DataLoader): returns Tuple[pd.DataFrame, pd.DataFrame]: prediction values, target values
                        evaluate(DataLoader): returns float: evaluation metric
            loss_func (Loss): One of pytorch's loss classes
            optimizer (Optimizer): One of pytorch's optimizer classes
            scheduler (Scheduler): One of pytorch's scheduler classes
            class_independent (bool): If true, perform pruning and policy updates independently per output.
                Used in most multi-output cases.
            epochs (int): Maximum number of epochs to train for
            l1_lambda (float): L1 regularization weight
            early_stopping_plateau_count (int): Number of epochs without improvement to end training.
            perform_prune_plateau_count (int): Number of epochs without improvement to prune and grow model.
            accumulation_steps (int): Number of optimization steps to perform before backward operation.
            objective (str): 'maximize' or 'minimize'.  Default is 'maximize'.
        """
        if class_independent:
            setattr(loss_func, "reduction", 'none')

        super(BoostedBanditNRNTrainer, self).__init__(
            model=model,
            loss_func=loss_func,
            optimizer=optimizer,
            scheduler=scheduler,
            epochs=epochs,
            accumulation_steps=accumulation_steps,
            objective=objective,
            l1_lambda=l1_lambda,
            lookahead_steps=lookahead_steps,
            lookahead_steps_size=lookahead_steps_size,
            augment=augment,
            augment_alpha=augment_alpha,
            class_independent=class_independent,
            early_stopping_plateau_count=early_stopping_plateau_count,
            perform_prune_plateau_count=perform_prune_plateau_count,
            increase_prune_plateau_count=increase_prune_plateau_count,
            increase_prune_plateau_count_plateau_count=increase_prune_plateau_count_plateau_count
        )
        self.boosting_epochs = boosting_epochs
        self.partial_fit = partial_fit

    def boost(self, train_dl: DataLoader):
        """
        Train the rn model.

        Args:
            train_dl (DataLoader): data loader for training data
            val_dl (DataLoader): data loader for a held-out set used for early stopping
            evaluation_metric (function): scikit-learn evaluation metrics
            multi_class (bool): treat problem as multi-class for evaluation purposes

                Format:
                    DataLoaders must return batches with the keys:
                        'features': [BATCH_SIZE, N_FEATURES]
                        'target': [BATCH_SIZE, N_TARGETS]
        """
        partial_model = None
        all_features = []
        all_boosted_targets = []
        for epoch in range(self.boosting_epochs):
            for i, batch in enumerate(train_dl):
                # [BATCH_SIZE, N_FEATURES]
                features = batch['features']
                # [BATCH_SIZE, N_TARGETS]
                target = batch['target'].squeeze()

                if self.USE_CUDA:
                    features = features.cuda()
                    target = target.cuda()
                elif self.USE_MPS:
                    features = features.to('mps')
                    target = target.to('mps')

                # [BATCH_SIZE, N_TARGETS]
                predictions = self.model.rn(features)
                predictions = predictions.squeeze(-1)

                boosted_target = torch.abs(predictions - target)
                boosted_target = torch.where(torch.tensor(predictions > target), -1.0 * boosted_target, boosted_target)
                boosted_target = boosted_target.cpu().detach().numpy()
                features = features.cpu().detach().numpy()

                if self.partial_fit:
                    self.model.xgb.fit(features, boosted_target, xgb_model=partial_model)
                    partial_model = self.model.xgb
                else:
                    all_features += [features]
                    all_boosted_targets += [boosted_target]
        if not self.partial_fit:
            all_features = np.vstack(all_features)
            if self.model.output_size > 1:
                all_boosted_targets = np.vstack(all_boosted_targets)
            else:
                all_boosted_targets = np.concatenate(all_boosted_targets)
            self.model.xgb.fit(all_features, all_boosted_targets)

        self.model.xgb_is_fitted = True


__all__ = [BoostedBanditNRNTrainer]
